# Copyright(c) 2002 Association of Universities for Research in Astronomy, Inc.

procedure gdisplay(image,frame)

# Display GMOS (raw) images
#
# Version Feb 28, 2002   IJ  v1.3 release
#         Apr 26, 2002   IJ  change warning comment, handle [1] images
#         Jul  7, 2002   IJ  handle dir-names+short image names correctly
#         Sept 20, 2002  IJ   v1.4 release

char image      {"",prompt="GMOS image to display"}
int  frame      {1,prompt="Frame to write to"}
bool fl_paste   {no,prompt="Paste images to one for imexamine"}
bool fl_bias    {no,prompt="Rough bias subtraction"}
int  gap        {37,prompt="Size of the gap between the CCDs (in pixels)"}
real z1         {0.,prompt="Lower limit if not autoscaling"}
real z2         {0.,prompt="Upper limit if not autoscaling"}
char sci_ext   {"all",prompt="Name of extension(s) to display"}
char key_detsec  {"DETSEC",prompt="Header keyword for detector section"}
char key_datasec  {"DATASEC",prompt="Header keyword for data section"}
char key_ccdsum  {"CCDSUM",prompt="Header keyword for CCD binning"}
bool   verbose   {no,prompt="Verbose"}
struct* scanfile {"",prompt="Internal use only."}

begin

char l_image, l_sci_ext, l_key_detsec, l_key_datsec, l_key_ccdsum
bool l_verbose, l_fl_paste, l_fl_bias
int l_gap, l_frame
real l_z1, l_z2

char tmpin, l_dsec, l_datsec, l_stdimage, junk, tmpout, l_paste
char tmpim1, tmpim2, tmpim3, l_imext, tmpdisp
char l_pixtype
int n_ext, Xmin, Xmax, Ymin, Ymax, x1, x2, y1, y2, n_i, n_j, n_k
int Xoff, Yoff, Xbin, Ybin
real l_xcenter, l_ycenter, l_mag
bool fl_erase, fl_zscale, fl_zrange, l_fl_mag, l_fl_chmag, l_fl_sec
struct l_struct
real l_bias1[2], l_bias2[2], l_bias3[2]
int  l_igain

l_image=image ; l_sci_ext=sci_ext ; l_key_detsec=key_detsec
l_key_datsec=key_datasec ; l_key_ccdsum=key_ccdsum ; l_z1=z1 ; l_z2=z2
l_verbose=verbose ; l_gap=gap ; l_frame=frame ; l_fl_paste=fl_paste
Xmin=1 ; Xmax=6144+2.*l_gap; Ymin=1; Ymax=4608
l_fl_bias=fl_bias

# slow gain=2 bias values
l_bias1[1]=750.  # 790. in science exposures?
l_bias2[1]=630.
l_bias3[1]=387.

# fast gain=6 bias values
l_bias1[2]=657.
l_bias2[2]=534.
l_bias3[2]=350.


tmpin="uparm$"//mktemp("tmpin")
tmpout=mktemp("tmpout")
tmpdisp=mktemp("tmpdisp")
tmpim1=mktemp("tmpim1")
tmpim2=mktemp("tmpim2")
tmpim3=mktemp("tmpim3")

fl_zscale=yes ; fl_zrange=yes
if(l_z1!=0. || l_z2!=0.) {
  fl_zscale=no ; fl_zrange=no
}

cache("imgets","gimverify","fparse")
date | scan(l_struct)

gimverify(l_image)
if(gimverify.status==1) {
  print("ERROR - GDISPLAY: Cannot access image "//l_image)
  bye
}
if(gimverify.status!=0) {
  print("ERROR - GDISPLAY: Image "//l_image//" is not a MEF file")
  bye
}
fparse(gimverify.outname)
l_imext=fparse.root ; tmpout=tmpout//l_imext
if(fl_paste)
print("Temporary image name "//tmpout)

show stdimage | scan(l_stdimage)
printf("stdimage = %s\n",l_stdimage)
if(substr(l_stdimage,1,7)=="imtgmos" && l_stdimage!="imtgmosccd") {
#  l_imX=6400 ; l_imY=4644
  Xoff=nint(6400.-Xmax)/2.
  Yoff=nint(4644.-Ymax)/2.
  Xmax=6400 ; Ymax=4644
} else {
 print("ERROR - GDISPLAY: Please set stdimage to a full display GMOS device")
 print("                  imtgmos, imtgmos2, imtgmos4 or imtgmos8")
 bye
}
l_mag=1.
if(strlen(l_stdimage)==8) 
  l_mag=1./real(substr(l_stdimage,8,8))

if(!l_fl_paste && l_fl_bias) 
  print("WARNING - GDISPLAY: Cannot bias subtract if not pasting")

if(l_sci_ext!="all") {
  fxhead(l_image) | match("IMAGE","STDIN",stop-) | \
    match(l_sci_ext,"STDIN",stop-, > tmpin)
} else {
  n_k=0
  fxhead(l_image) | match("IMAGE","STDIN",stop-) | \
    match("SCI","STDIN",stop-) | scan(n_k)
  if(n_k>0) {
  fxhead(l_image) | match("IMAGE","STDIN",stop-) | \
    match("SCI","STDIN",stop-, > tmpin)
  print("Displaying the SCI extensions")
  } else {
    fxhead(l_image) | match("IMAGE","STDIN",stop-, > tmpin)
  }
}

count(tmpin) | scan(n_ext)  # number of extensions

delete("uparm$gdisplay"//str(l_frame),verify-, >>& "dev$null")
delete("uparm$mscdisp"//str(l_frame),verify-, >>& "dev$null")

l_fl_chmag=yes  # check magnification on first round
l_fl_mag=no     # initialize check for changed magnification

n_k=0 ; l_fl_sec=yes
scanfile=tmpin
while(fscan(scanfile,n_i)!=EOF) {
 n_k=n_k+1
# Get binning
 imgets(l_image//"["//str(n_i)//"]",l_key_ccdsum)
 if(imgets.value=="0") {
   Xbin=1 ; Ybin=1
 } else {
   print(imgets.value) | \
    fields("STDIN","1,2",lines="1",quit_if_miss-,print_file-) | \
    scan(Xbin,Ybin)
 }

 imgets(l_image//"["//str(n_i)//"]",l_key_detsec, >>& "dev$null")
 l_dsec=imgets.value
 imgets(l_image//"["//str(n_i)//"]",l_key_datsec, >>& "dev$null")
 l_datsec=imgets.value
if(l_dsec=="0" || l_datsec=="0") 
 l_fl_sec=no

# Get the readout mode  (slow/fast lowgain modes only)
 if(n_k==1) {
 imgets(l_image//"[0]","AMPINTEG", >>& "dev$null")
 if(imgets.value=="1000")
   l_igain=2
  else
   l_igain=1
 }

if((l_fl_paste || l_fl_bias) && !l_fl_sec) {
  print("WARNING - GDISPLAY: No data section info found, or image is already mosaiced")
  print("                    Displaying input image. Using fl_paste=no fl_bias=no")
  l_fl_paste=no ; l_fl_bias=no
}

if(!l_fl_paste) {

 if(l_fl_sec) {
# find x1, x2, y1, y2 and set display parameters
 n_j=stridx(":",l_dsec)
 x1=int(substr(l_dsec,2,n_j-1))
 l_dsec=substr(l_dsec,n_j+1,strlen(l_dsec))
 n_j=stridx(",",l_dsec)
 x2=int(substr(l_dsec,1,n_j-1))
 l_dsec=substr(l_dsec,n_j+1,strlen(l_dsec))
 n_j=stridx(":",l_dsec)
 y1=int(substr(l_dsec,1,n_j-1))
 l_dsec=substr(l_dsec,n_j+1,strlen(l_dsec))
 y2=int(substr(l_dsec,1,strlen(l_dsec)-1))
 } else {
    x1=1 ; y1=1
    imgets(l_image//"["//str(n_i)//"]","i_naxis1")
    x2=int(imgets.value)*Xbin
    imgets(l_image//"["//str(n_i)//"]","i_naxis2")
    y2=int(imgets.value)*Ybin
 }

 printf("%s %4d %4d %4d %4d\n",l_image//"["//str(n_i)//"]",x1,x2,y1,y2, >> "uparm$gdisplay"//str(l_frame))
 printf("%s %4d %4d %4d %4d\n",l_image//"["//str(n_i)//"]",x1,x2,y1,y2, >> "uparm$mscdisp"//str(l_frame))

if(n_k==1)
 fl_erase=yes
else
 fl_erase=no

l_xcenter=(x2-x1)/2.+x1
if(l_xcenter>2048)
  l_xcenter+=l_gap
if(l_xcenter>(2*2048.+l_gap))
  l_xcenter+=l_gap
l_xcenter+=Xoff

l_xcenter=l_xcenter/Xmax
l_ycenter=((y2-y1)/2.+y1+Yoff)/Ymax
#print(l_xcenter,l_ycenter)

# Reset xcenter,ycenter if image is binned
  l_xcenter = (l_xcenter-0.5)/Xbin + 0.5
  l_ycenter = (l_ycenter-0.5)/Ybin + 0.5

# Reset magnification if higher resolution can be displayed
# Never display more than 1pix to 1pix
  if(min(Xbin,Ybin)>1. && l_mag*min(Xbin,Ybin) <= 1. && l_fl_chmag) {
     l_mag=l_mag*min(Xbin,Ybin)
     l_fl_chmag=no  # don't check/change l_mag in the next rounds
     l_fl_mag=yes   # update xcenter,ycenter in all rounds
  }
# Adjust xcenter,ycenter if l_mag has been changed
  if(l_fl_mag) {
     l_xcenter = (l_xcenter-0.5)*min(Xbin,Ybin) + 0.5
     l_ycenter = (l_ycenter-0.5)*min(Xbin,Ybin) + 0.5
  } 

 printf("%s %4d %4d %4d %4d  %5.1f ",l_image//"["//str(n_i)//"]",x1,x2,y1,y2,l_mag)
 display(l_image//"["//str(n_i)//"]"//l_datsec,l_frame,erase=fl_erase,
  xcenter=l_xcenter,ycenter=l_ycenter,z1=l_z1,z2=l_z2,zrange=fl_zrange,
  zscale=fl_zscale,xmag=l_mag,ymag=l_mag,fill=no)

} else if (l_fl_sec) {
# Setup parameters for imtile
  if(n_k==1) {
    l_gap=nint(l_gap/Xbin)
    l_paste=l_image//"["//str(n_i)//"]"//l_datsec
    imgets(l_image//"["//str(n_i)//"]","i_pixtype")
    if(imgets.value=="11")
       l_pixtype="l"
    else
       l_pixtype="r"

  } else
    l_paste=l_paste//","//l_image//"["//str(n_i)//"]"//l_datsec

}


}
scanfile=""

delete(tmpin,verify-, >>& "dev$null")

if(l_fl_paste) {

 print("GDISPLAY: Pasting images - currently only correct for 3amps frames")
 if(l_fl_bias && n_ext==3) {
  imarith(l_paste,"-",
    str(l_bias1[l_igain])//","//str(l_bias2[l_igain])//","//str(l_bias3[l_igain]),
    tmpim1//","//tmpim2//","//tmpim3,
    pixtype=l_pixtype,calctype=l_pixtype,verbose-,noact-,divzero=0.,title="",
    hparams="")
  l_paste=tmpim1//","//tmpim2//","//tmpim3
 } else if (l_fl_bias && n_ext!=3) {
  print("WARNING - GDISPLAY: Currently only 3amps full frames can be bias subtracted")
 }

 imtile(l_paste,tmpout,n_k,1,ncoverlap=-l_gap,trim_section="",
   missing_inpu="",start_tile="ll",row_order=yes,raster_order=no,
   median_secti="",subtract=no,ncols=INDEF,nlines=INDEF,nloverlap=-1,
   opixtype=l_pixtype,ovalue=0,verbose=l_verbose)
 display(tmpout,l_frame,erase=yes,xcenter=0.5,ycenter=0.5,
    xmag=1,ymag=1,z1=l_z1,z2=l_z2,zrange=fl_zrange,zscale=fl_zscale,
    fill=no)
 print("GDISPLAY: Starting imexamine - quit with q")
 imexamine()
 imdelete(tmpout,verify-, >>& "dev$null")
 if(l_fl_bias && l_fl_sec && n_ext==3)
   imdelete(l_paste,verify-)

}

end
