# Copyright(c) 2002 Association of Universities for Research in Astronomy, Inc.

procedure gfdisplay (input)

# Display extracted GMOS IFU spectra using ldisplay
# The scale of the output image is 0.0125arcsec/pixel
#
# Version   Sept 20, 2002 BM  v1.4 release

string input      {"",prompt="Input GMOS IFU extracted spectra"}
int    frame      {1,prompt="Frame for display"}
real   z1         {0.0,prompt="Minimum greylevel to be displayed"}
real   z2         {0.0,prompt="Maximum greylevel to be displayed"}
string ztrans     {"log",enum="linear|log",prompt="Greylevel transformation (linear|log)"}
string output     {"",prompt="Name of output image"}
string extname    {"SCI",prompt="Which EXTNAME to display"}
string version    {"*",enum="1|2|*",prompt="Which EXTVER to display (1,2,*)"}
string config     {"gmos$data/gnifu_slits.cfg",prompt="Ldisplay configuation file"}
string deadfib    {"gmos$data/gnifu_deads.cfg",prompt="Ldisplay dead fiber configuation file"}
int    status     {0,prompt="Exit status (0=good)"}

begin

string l_input,l_config, l_ver, l_extn, l_dead,l_output,l_ztrans
int l_frame
real l_z1,l_z2

bool rmout
int nx1,nx2,len
real ref1,ref2,refpix
string tmpmsa,tmpmsb,msjoin,tmpout
string sec1,sec2,cmd

l_input=input; l_frame=frame; l_extn=extname; l_ver=version
l_config=config; l_dead=deadfib; l_output=output; l_z1=z1; l_z2=z2
l_ztrans=ztrans

cache("imgets")

tmpmsa=mktemp("tmpmsa")
tmpmsb=mktemp("tmpmsb")
msjoin=mktemp("tmpmsjoin")
tmpout=mktemp("tmpldisplay")
rmout=no
status=0

# check that there are input files
if (l_input == "" || l_input == " "){
  print("ERROR - GFDISPLAY: input files not specified")
  goto error
}

# check existence of list file
if (substr(l_input,1,1) == "@") {
    print("ERROR - GFDISPLAY: lists are currently not supported")
    goto error
}

# check existence of input file
gimverify(l_input)
if (gimverify.status != 0) {
  print("ERROR - GFDISPLAY: "//l_input//" does not exist or is not a MEF")
  goto error
}
l_input=gimverify.outname//".fits"

# check existence of config files
if (!access(l_config)) {
  print("ERROR - GFDISPLAY: "//l_config//" does not exist.")
  goto error
}
if (!access(l_dead)) {
  print("ERROR - GFDISPLAY: "//l_dead//" does not exist.")
  goto error
}

# Command for display
cmd="display frame="//str(l_frame)
if (l_z1 != 0.0 || l_z2 != 0.0) {
    cmd=cmd//" zscale- zrange- z1="//str(l_z1)//" z2="//str(l_z2)
}
cmd=cmd//" ztrans="//l_ztrans//" image="

# Temporary output output image if no output image given
if (l_output=="" || l_output==" ") {
   l_output=tmpout
   rmout=yes
}

if (l_ver == "1" || l_ver=="2") {
    if (!imaccess(l_input//"["//l_extn//","//l_ver//"]")) {
        print("ERROR - GFDISPLAY: "//l_input//"["//l_extn//","//l_ver//"] does not exist.")
        goto error
    }
    imcopy(l_input//"["//l_extn//","//l_ver//"]",msjoin//".ms",verbose-)
} else if (l_ver=="*") {
    if (!imaccess(l_input//"["//l_extn//",1]")) {
        print("ERROR - GFDISPLAY: "//l_input//"["//l_extn//",1] does not exist.")
        goto error
    }
    if (!imaccess(l_input//"["//l_extn//",2]")) {
        print("ERROR - GFDISPLAY: "//l_input//"["//l_extn//",2] does not exist.")
        goto error
    }
    imgets(l_input//"["//l_extn//",1]","i_naxis1", >& "dev$null")
    nx1=int(imgets.value)
    imgets(l_input//"["//l_extn//",1]","refpix1", >& "dev$null")
    ref1=real(imgets.value)
    imgets(l_input//"["//l_extn//",2]","i_naxis1", >& "dev$null")
    nx2=int(imgets.value)
    imgets(l_input//"["//l_extn//",2]","refpix1", >& "dev$null")
    ref2=real(imgets.value)

    len=min(nx1,nx2)
    refpix=min(ref1,ref2)
    sec1="["//(nint(ref1-refpix+1.))//":"//(nint(ref1-refpix+len))//",*]"
    sec2="["//(nint(ref2-refpix+1.))//":"//(nint(ref2-refpix+len))//",*]"
    #print(sec1, sec2)

    imcopy(l_input//"["//l_extn//",1]"//sec1,tmpmsa//".ms",verb-)
    imcopy(l_input//"["//l_extn//",2]"//sec2,tmpmsb//".ms",verb-)
    scopy(tmpmsa//".ms,"//tmpmsb//".ms",msjoin//".ms",renum+,merge+,clobber+,
    verb-,rebin-)
}
count(l_config) | scan(nx2)
imgets(msjoin//".ms.fits","i_naxis2")
if (nx2==int(imgets.value)) {
    ldisplay(msjoin//".ms",siz=1024,ldispdir="",lconf=l_config,lscale=0.2,
    deadfibs=l_dead,output=l_output,axisflip+,disqu=yes,display=cmd)
} else {
    print("ERROR - GFDISPLAY: ldisplay configuration file does not match image")    
}

# clean up
goto clean

error:
    status=1
    goto clean

clean:
    imdelete(tmpmsa//".ms,"//tmpmsb//".ms,"//msjoin//".ms",verify-, >>& "dev$null")
    if (rmout) {
        imdelete(l_output,verify-, >>& "dev$null")
    }

end
