# Copyright(c) 2002 Association of Universities for Research in Astronomy, Inc.

procedure gfextract (inimage)

# Extract GMOS IFU spectra
#
# Version   Sept 20, 2002 BM v1.4 release

string inimage   {"",prompt="Input image"}
string outimage  {"",prompt="Output file"}
string outpref   {"e",prompt="Prefix for output image"}
string title     {"",prompt="Title for output SCI plane"}
string reference {"",prompt="Reference file"}
string response  {"",prompt="Fiber response file"}
int    line      {INDEF,prompt="Line/column for finding apertures"}
int    nsum      {10,min=1,prompt="Number of columns to use for finding apertures"}
bool   trace     {yes,prompt="Trace spectra?"}
bool   recenter  {yes,prompt="Recenter apertures?"}
string function  {"chebyshev",enum="chebyshev|spline1|spline3|legendre",prompt="Function for trace"}
int    order     {5,min=1,prompt="Order of trace fit"}
int    t_nsum    {10,min=1,prompt="Number of dispersion lines to sum for the trace"}
string weights   {"variance",enum="variance|none",prompt="Weighting during extraction"}
string bpmfile   {"gmos$data/chipgaps.dat",prompt="Bad pixel mask for column interpolation"}
string gratingdb  {"gmos$data/GMOSgratings.dat",prompt="Gratings database file"}
string filterdb   {"gmos$data/GMOSfilters.dat",prompt="Filters database file"}
real   xoffset    {INDEF,prompt="X offset in wavelength [nm]"}
string sci_ext    {"SCI",prompt="Name of science extension"}
string var_ext    {"VAR",prompt="Name of variance extension"}
string dq_ext     {"DQ",prompt="Name of data quality extension"}
bool   fl_vardq   {no,prompt="Extract variance and DQ planes?"}
bool   fl_gsappwave {yes,prompt="Run GSAPPWAVE?"}
bool   fl_inter   {no,prompt="Run APALL interatively?"}
string logfile    {"",prompt="Logfile"}
bool   verbose    {yes,prompt="Verbose output?"}
int    status     {0,prompt="Exit status (0=good)"}

begin

string l_inimage,l_outimage,l_prefix,l_title,l_logfile,l_response
string l_grating,l_filter,l_gratingdb,l_filterdb,l_reference,l_bpmfile
string l_function,l_weights
string l_sci_ext,l_var_ext,l_dq_ext
string imgsec,slit[2],mdf,impaste
string sci,var,dq,aptable,mdfslit
string gname,fname,ffile,outroot,refroot,outlog
int l_xbin,l_ybin,l_line,l_nsum,l_t_nsum,l_order
int lmm,sx,sy,wx,width,center,xccd[2],yccd[2]
int sposy,slitsepp
real l_wave,blambda,R,coverage,nmppx,scale,asecmm,l_xoffset
real gwave1,gwave2,fwave1,fwave2,wave1,wave2,wmin,wmax,wavoffset
real tilt,pi,a,greq,refpix[2],refmin
int nslit,j1,id1,nr,x1[2],x2[2],j,len,nx[2]
int nextnd,nfind[2]
bool l_verbose,l_inter,l_recenter,l_trace,delpaste, l_fl_vardq
bool l_fl_gsappwave,l_fl_fixpix
struct sdate

# Query parameters
l_inimage=inimage ; l_outimage=outimage ; l_title=title ; l_prefix=outpref
l_response=response ; l_reference=reference ; l_bpmfile=bpmfile
l_line=line; l_logfile=logfile; l_verbose=verbose; l_inter=fl_inter
l_fl_vardq=fl_vardq; l_gratingdb=gratingdb; l_filterdb=filterdb
l_trace=trace; l_nsum=nsum; l_t_nsum=t_nsum; l_order=order
l_function=function; l_weights=weights; l_fl_gsappwave=fl_gsappwave
l_recenter=recenter
l_sci_ext=sci_ext ; l_var_ext=var_ext ; l_dq_ext=dq_ext
l_xoffset=xoffset

status=0
pi=3.1415927
scale=0.0727
asecmm=1.611444
delpaste=yes
nfind[1]=749 ; nfind[2]=747

# Keep imgets parameters from changing by outside world
cache("imgets", "hedit", "fparse","tinfo")

# Define temporary files
imgsec=mktemp("tmpimgsec")
sci=mktemp("tmpsci")
var=mktemp("tmpvar")
dq=mktemp("tmpdq")
mdf=mktemp("tmpmdf")
impaste=mktemp("tmppaste")
aptable=mktemp("tmpaptable")
mdfslit=mktemp("tmpmdfslit")

# Start logging to file
if (l_logfile == "STDOUT") {
   l_logfile = ""
   l_verbose=yes
}

# Test the logfile:
gemlogname(logpar=l_logfile,package="gmos")
if (gemlogname.status != 0) {
    goto error
}
l_logfile=gemlogname.logname

# Start logfile
date | scan(sdate)
printlog("--------------------------------------------------------------------------------",l_logfile,verbose=l_verbose)
printlog("GFEXTRACT -- "//sdate,l_logfile,verbose=l_verbose)
printlog(" ",l_logfile,verbose=l_verbose)
printlog("inimage   = "//l_inimage,l_logfile,verbose=l_verbose)
printlog("outimage  = "//l_outimage,l_logfile,verbose=l_verbose)
printlog("outpref   = "//l_prefix,l_logfile,verbose=l_verbose)
printlog("title     = "//l_title,l_logfile,verbose=l_verbose)
printlog("response  = "//l_response,l_logfile,verbose=l_verbose)
printlog("reference = "//l_reference,l_logfile,verbose=l_verbose)
printlog("trace     = "//l_trace,l_logfile,verbose=l_verbose)
printlog("function  = "//l_function,l_logfile,verbose=l_verbose)
printlog("order     = "//l_order,l_logfile,verbose=l_verbose)
printlog("weights   = "//l_weights,l_logfile,verbose=l_verbose)
printlog("bpmfile   = "//l_bpmfile,l_logfile,verbose=l_verbose)
printlog("gratingdb = "//l_gratingdb,l_logfile,verbose=l_verbose)
printlog("filterdb  = "//l_filterdb,l_logfile,verbose=l_verbose)
printlog("xoffset   = "//str(l_xoffset),l_logfile,verbose=l_verbose)
printlog("fl_vardq  = "//l_fl_vardq,l_logfile,verbose=l_verbose)
printlog(" ",l_logfile,verbose=l_verbose)

#check that there are input files
if (l_inimage == "" || l_inimage == " "){
  printlog("ERROR - GFEXTRACT: input files not specified",l_logfile,yes)
  goto error
}

# check existence of list file
if (substr(l_inimage,1,1) == "@") {
    printlog("ERROR - GFEXTRACT: lists are currently not supported",l_logfile,yes)
    goto error
}

# check existence of input file
gimverify(l_inimage)
if (gimverify.status != 0) {
  printlog("ERROR - GFEXTRACT: "//l_inimage//" does not exist or is not a MEF",l_logfile,yes)
  goto error
}
l_inimage=gimverify.outname//".fits"

#check that an output file is given
if ((l_outimage == "" || l_outimage == " ") && (l_prefix=="" || l_prefix==" ")) {
    printlog("ERROR - GFEXTRACT: output file not specified",l_logfile,yes)
    goto error
}

# Check if using prefix
if (l_outimage == "" || l_outimage == " ") {
    l_outimage=l_prefix//l_inimage
}

#check output name
fparse(l_outimage,verbose-)
outroot=fparse.root
if (fparse.extension == "") {
    l_outimage=l_outimage//".fits"
}

# check response file if not ""
if (l_response != "" && l_response != " ") {
    gimverify(l_response)
    if (gimverify.status != 0) {
        printlog("ERROR - GFEXTRACT: response image does not exist or is not a MEF.",l_logfile,yes)
        goto error
    }
    l_response=gimverify.outname//".fits"
    imgets(l_response//"[0]","GFRESPON", >>& "dev$null")
    if (imgets.value == "0") {
        printlog("ERROR - GFEXTRACT: response image has not been processed by GFRESPONSE.",l_logfile,yes)	
        goto error
    }   
} else {
    l_response=""
}

# check bpmfile file if not ""
if (l_bpmfile != "" && l_bpmfile != " ") {
    if (!access(l_bpmfile)) {
        printlog("ERROR - GFEXTRACT: bpmfile image does not exist.",l_logfile,yes)
        goto error
    }
    l_fl_fixpix=yes
} else {
    l_fl_fixpix=no
}

# reference
fparse(l_reference,verbose-)
refroot=fparse.root

# Check that the output file does not already exist. If so, exit.
if (imaccess(l_outimage)) {  
  printlog("ERROR - GFEXTRACT: Output file "//l_outimage//" already exists.",l_logfile,yes)
  goto error
}

# slice out the MDF
tcopy(l_inimage//"[MDF]",mdf//".fits",verbose-)
if (!access(mdf//".fits")) {
    printlog("ERROR - GFEXTRACT: Input image does not contain a MDF",l_logfile,yes)
    goto error
}

# Run GGAIN if not done already
imgets(l_inimage//"[0]","GGAIN", >>& "dev$null")
if (imgets.value=="0") {
    ggain(l_inimage,gaindb="gmos$data/gmosamps.dat",logfile=l_logfile,
    key_gain="GAIN",key_ron="RDNOISE",fl_mult+,fl_update+,sci_ext=l_sci_ext,
    var_ext=l_var_ext,verbose=l_verbose)
    if (ggain.status != 0) {
        goto error
    }
}
 
# Find slits
# get header information
imgets(l_inimage//"[0]","GRATING")
l_grating=imgets.value
if (l_grating=="MIRROR") {
    printlog("ERROR - GFEXTRACT: input image is not dispersed.",l_logfile,l_verbose)
    goto error
}

imgets(l_inimage//"[sci,1]","CCDSUM") ; print(imgets.value) | scan(l_xbin,l_ybin)
imgets(l_inimage//"[0]","FILTER2") ; l_filter=imgets.value
imgets(l_inimage//"[0]","GRWLEN") ; l_wave=real(imgets.value)

# get grating information
match(l_grating,l_gratingdb,stop-,print+,meta+) | scan(gname,lmm,blambda,R,coverage,gwave1,gwave2,wavoffset)

if (l_xoffset != INDEF) 
    wavoffset=l_xoffset

#l_wave=l_wave+wavoffset
print(((l_wave*lmm)/1.e6)) | interp("gmos$data/gratingeq.dat","STDIN",int_mode="spline",curve_gen-) | scan(greq,tilt)

tilt=tilt*pi/180.
a=sin(tilt+0.872665)/sin(tilt)
R=206265.*greq/(0.35*81.0*sin(tilt))
nmppx=a*scale*real(l_xbin)*l_wave/(R*0.35)
coverage=nmppx*6218./real(l_xbin)
slitsepp=175.*asecmm/(scale*a)

wave1=gwave1
wave2=gwave2

printlog("Grating: "//gname,l_logfile,l_verbose)
printlog("nm/pix = "//nmppx,l_logfile,l_verbose)
printlog("Central wavelength = "//l_wave,l_logfile,l_verbose)

# which slit?
tinfo(mdf//".fits",ttout-)
nr=tinfo.nrows
tabpar(mdf//".fits","NO",1)
id1=int(tabpar.value)
nslit=1
if (nr>750) {
    nslit=2
}
j1=1
if (id1==751) {
    j1=2
    nslit=2
}

# get filter information
fwave1=0.0
fwave2=9999.0
if (l_filter != "" && l_filter != "open2-6") {
    match(l_filter,l_filterdb,stop-,print+,meta+) | scan(fname,fwave1,fwave2,ffile)
}

# determine whether filter or grating limits wavelength coverage
if (fwave1 > wave1) {
    wave1 = fwave1
}
if (fwave2 < wave2) {
    wave2 = fwave2
}
# and red limit of CCD
if (wave2 > 1050.) {
    wave2 = 1050.
}

# length of spectrum in x
# Maximum length possible with grating coverage
wmin=wave1
wmax=wave2

# in pixels
sx=nint((wmax-wmin)/nmppx)
printlog("Filter: "//l_filter,l_logfile,l_verbose)
printlog("Max wavelength coverage: "//wmin//" "//wmax,l_logfile,l_verbose)
#printlog("Max spectrum length = "//sx,l_logfile,l_verbose)

# pixel value of central wavelength from left (red) end of spectrum
# This is not the grating configured central wavelength, but the center 
# wavelength of the possible spectral coverage 
wx=sx-nint((l_wave-wmin)/nmppx)
#printlog("Pixel of central wavelength: "//wx,l_logfile,l_verbose)

# positions of slits on CCD
yccd[1]=1
yccd[2]=1
xccd[1]=(3109-slitsepp/2)/l_xbin+wavoffset/nmppx
xccd[2]=xccd[1]+slitsepp/l_xbin
printlog("Slit separation: "//slitsepp,l_logfile,l_verbose)
printlog("Positions of slits: "//xccd[1]//" "//xccd[2],l_logfile,l_verbose)

if (sx > slitsep && nslit-j1==1) {
    sx=slitsepp
}

# slice out the two slits
for (i=1; i<=2; i+=1) {
    refpix[i]=real(wx)
    x1[i]=xccd[i]-wx
    if (x1[i] < 1) {
        refpix[i]=refpix[i]-real(-x1[i]+1)
        x1[i]=1
    }
    x2[i]=xccd[i]-wx+sx
    if (x2[i] > 6218/l_xbin) {
# deal correctly with binned data
        if(l_xbin==1)
        x2[i]=6218/l_xbin
        else
        x2[i]=6216/l_xbin
    }
    nx[i]=x2[i]-x1[i]+1.
}
# Extract spectra of the same length from each slit
refmin=min(refpix[1],refpix[2])
if (nslit-j1==1) {
# deal correctly with binned data
    if(l_xbin==1)
    len=min(nx[1],(refmin+(6218/l_xbin-xccd[2])))
    else
    len=min(nx[1],(refmin+(6216/l_xbin-xccd[2])))

    slit[1]="["//(nint(xccd[1]-refmin+1.))//":"//(nint(xccd[1]-refmin+len))//",*]"
    slit[2]="["//(nint(xccd[2]-refmin+1.))//":"//(nint(xccd[2]-refmin+len))//",*]"
} else {
    for (i=1; i<=2; i+=1) {
        slit[i]="["//x1[i]//":"//(x2[i])//",*]"
    }
}

printlog("Slit section(s): "//slit[1]//" "//slit[2],l_logfile,l_verbose)
printlog("",l_logfile,l_verbose)

# Paste the MEF together
imgets(l_inimage//"[0]","GMOSAIC", >& "dev$null")
if (imgets.value == "0" && l_outimage!="") {
    printlog("Mosaicing",l_logfile,l_verbose)
    gmosaic(l_inimage,outimages=impaste,gap=37,fl_paste-,sci_ext=l_sci_ext,
    key_detsec="DETSEC",key_datsec="DATASEC",key_ccdsum="CCDSUM",
    logfile=l_logfile,verbose=l_verbose,fl_fixpix=l_fl_fixpix,
    bpmfile=l_bpmfile)
    delpaste=yes
} else {
    impaste=l_inimage
    delpaste=no
}

# Copy MDF to output image
wmef(mdf//".fits",l_outimage,extnames="MDF",verb-,phu=impaste, >& "dev$null")

if (l_title != "" && l_title != " ") {
    hedit(l_outimage//"[0]","i_title",l_title,add+,update+,verify-,show-)
} else {
    imgets(l_inimage//"[0]","i_title", >>& "dev$null")
    if (imgets.value != "0") {
        l_title=imgets.value
    }
}

# Loop over slits
outlog="dev$null"
if (l_inter) {
    outlog="STDOUT"
}
 
j=0
for (i=j1; i<=nslit; i+=1) {
    printlog("",l_logfile,l_verbose)
    printlog("Extracting slit "//i,l_logfile,l_verbose)
    j=j+1

    # Copy image section since apall won't accept section extensions
    imgsec=outroot//"_"//i
    if (imaccess(imgsec)) {
        imdelete(imgsec,verify-)
    }
    imcopy(impaste//"["//l_sci_ext//",1]"//slit[i],imgsec,verbose-)
    hedit(imgsec,"DISPAXIS",1,add+,update+,verify-,show-)

    # Reference
    #l_recenter=yes
    #l_trace=yes
    if (refroot != "") {
        l_reference=refroot//"_"//i
        if (!access("database/ap"//l_reference)) {
            printlog("ERROR - GFEXTRACT: Aperture reference database/ap"//l_reference//" not found.",l_logfile,yes)
            goto error
        }
        #l_trace=no
        #l_inter=no
        #if (l_line == INDEF) {
        #    l_recenter=no
        #}
    }

    # Make apidtable from MDF
    tselect(l_outimage//"[MDF]",mdfslit,"NO>="//(750*(i-1)+1)//" && NO<="//(750*i)//" && BEAM >= 0")
    tprint(mdfslit,prparam-,prdata+,showrow-,showhdr-,showunits-,
    columns="NO,BEAM,XINST,YINST,BLOCK",rows="-",
    option="plain",align+,sp_col="",lgroup=0, > aptable)
    tinfo(aptable,ttout-)
    # Apall
    apall(imgsec,tinfo.nrows,output="tmpslit"//i//".ms",apertures="",
    format="multispec",references=l_reference,profiles="",inter=l_inter,
    find+,recenter=l_recenter,resize-,edit+,trace=l_trace,fittrace+,
    extract+,extras=l_fl_vardq,
    review+,line=l_line,nsum=l_nsum,lower=-2.5,upper=2.5,apidtable=aptable,
    width=5.,radius=10.,threshold=10.,minsep=4.0,maxsep=200,
    order="increasing",aprecenter="",npeaks=INDEF,shift+,llimit=INDEF,
    ulimit=INDEF,ylevel=0.1,peak+,bkg-,r_grow=0.,avglimits-,
    t_nsum=l_t_nsum,t_step=l_t_nsum,t_nlost=4,t_function=l_function,t_order=l_order,
    t_sample="*",t_naverage=1,t_niterate=3,t_low_reject=3.,t_high_reject=3.,
    background="none",weights=l_weights,pfit="fit1d",clean-,saturation=INDEF,
    gain=1.,lsigma=4.,usigma=4.,nsubaps=1, >>& outlog)

    # check if output image exists, an interactive user may have quit or
    # apall may have crashed...
    if (!imaccess("tmpslit"//i//".ms.fits")) {
        printlog("ERROR - GFEXTRACT: apall output does not exist.",l_logfile,yes)
        goto error
    }

    # kluge the header
    hedit("tmpslit"//i//".ms.fits","CD1_1,CD2_2",delete+,update+,verify-,show-)
    hedit("tmpslit"//i//".ms.fits","CDELT*",delete+,update+,verify-,show-)
    hedit("tmpslit"//i//".ms.fits","CD1_1",1.0,add+,update+,verify-,show-)
    hedit("tmpslit"//i//".ms.fits","CD2_2",1.0,add+,update+,verify-,show-)
    hedit("tmpslit"//i//".ms.fits","CRPIX1",1,add+,update+,verify-,show-)
    hedit("tmpslit"//i//".ms.fits","CRPIX2",1,add+,update+,verify-,show-)
    hedit("tmpslit"//i//".ms.fits","CRVAL1",1.,add+,update+,verify-,show-)
    hedit("tmpslit"//i//".ms.fits","CRVAL2",1.,add+,update+,verify-,show-)

    printlog("Extraction done",l_logfile,l_verbose)

    # Separate the multispec file
    imcopy("tmpslit"//i//".ms.fits[*,*,1]",sci,verb-)
    if (l_fl_vardq) {
        imarith("tmpslit"//i//".ms.fits[*,*,3]","*","tmpslit"//i//".ms.fits[*,*,3]",var,verb-)
        imarith(sci,"*",0,dq,pixtype="int",calctype="int",verb-)
    }

    # Apply fiber throughput correction 
    if (l_response != "") {
        if (imaccess(l_response//"["//l_sci_ext//","//j//"]")) {
            printlog("Applying fiber response",l_logfile,l_verbose)
            imarith(sci,"/",l_response//"["//l_sci_ext//","//j//"]",sci,verb-,noact-)
        } else {
            printlog("GFEXTRACT - ERROR: "//l_response//"["//l_sci_ext//","//j//"] not found",l_logfile,yes)
            goto error
        }
    }

    # Write to output
    printlog("Writing to output file",l_logfile,l_verbose)
    imgets(l_outimage//"[0]","NEXTEND")
    nextnd=int(imgets.value)
    if (l_fl_vardq) {
        fxinsert(sci//".fits,"//var//".fits,"//dq//".fits",l_outimage//"["//nextnd//"]","0",verbose-, >& "dev$null")
        hedit(l_outimage//"[0]","NEXTEND",(nextnd+3),update+,verify-,show-)
    } else {
        fxinsert(sci//".fits",l_outimage//"["//nextnd//"]","0",verbose-, >& "dev$null")
        hedit(l_outimage//"[0]","NEXTEND",(nextnd+1),update+,verify-,show-)
    }
    hedit(l_outimage//"["//(nextnd+1)//"]","EXTNAME",l_sci_ext,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+1)//"]","EXTVER",j,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+1)//"]","i_title",l_title,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+1)//"]","IFUSLIT",i,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+1)//"]","REFPIX1",refmin,add+,update+,verify-,show-)
    if (l_fl_vardq) {
    hedit(l_outimage//"["//(nextnd+2)//"]","EXTNAME",l_var_ext,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+2)//"]","EXTVER",j,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+2)//"]","i_title","Variance: "//l_title,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+2)//"]","IFUSLIT",i,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+2)//"]","REFPIX1",refmin,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+3)//"]","EXTNAME",l_dq_ext,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+3)//"]","EXTVER",j,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+3)//"]","i_title","Data Quality: "//l_title,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+3)//"]","IFUSLIT",i,add+,update+,verify-,show-)
    hedit(l_outimage//"["//(nextnd+3)//"]","REFPIX1",refmin,add+,update+,verify-,show-)
    }
    imdelete(imgsec//","//sci//","//var//","//dq,verify-, >& "dev$null")
    imdelete("tmpslit*.ms.fits",verify-, >& "dev$null")
    delete(aptable//","//mdfslit//".tab",verify-, >& "dev$null")
}

# final header update
hedit(l_outimage//"[0]","NSCIEXT",j,add+,update+,verify-,show-)
gemhedit(l_outimage//"[0]","GFEXTRAC",sdate,"Time stamp for GFEXTRACT")
gemhedit(l_outimage//"[0]","GEM-TLM",sdate,"Last modification with GEMINI")

# gsappwave
if (l_fl_gsappwave) {
    printlog("",l_logfile,l_verbose)
    gsappwave(l_outimage,logfile=l_logfile,gratingdb=l_gratingdb,
    filterdb=l_filterdb,key_dispaxis="DISPAXIS",dispaxis=1,verbose=l_verbose)
}

# clean up
goto clean

error:
    status=1
    goto clean

clean:
    delete("tmpslit*.fits,"//mdf//".fits"//","//aptable,verify-, >& "dev$null")
    imdelete(imgsec//","//sci//","//var//","//dq,verify-, >& "dev$null")
    delete(aptable//","//mdfslit//".tab",verify-, >& "dev$null")
    if (delpaste) {
        imdelete(impaste,verify-, >& "dev$null")
    }
    # close log file
    printlog(" ",l_logfile,l_verbose)
    if(status==0)
      printlog("GFEXTRACT exit status: good",l_logfile,l_verbose)
    else
      printlog("GFEXTRACT exit status: error",l_logfile,l_verbose)
    printlog("--------------------------------------------------------------------------------",l_logfile,l_verbose )

end

