# Copyright(c) 2002 Association of Universities for Research in Astronomy, Inc.

procedure gsflat (inflats,specflat,fl_answer)

# Create a GMOS normalized spectral flatfield (LONGSLIT or MOS) from GCAL flats
#
# Version    Feb 28, 2002  ML,BM,IJ  v1.3 release
#            Aug 15, 2002  BM  add x/yoffset parameters for new gscut
#            Sept 20, 2002    v1.4 release

string inflats       {"",prompt="Input flatfields"} 
string specflat      {"",prompt="Output normalized flat (MEF)"}
bool   fl_slitcorr   {no,prompt="Correct output for Illumination/Slit-Function?"}
string slitfunc      {"",prompt="Slit Function (MEF output of gsslitfunc)"} 
bool   fl_keep       {no,prompt="Keep imcombined flat?"} 
string combflat      {"",prompt="Filename for imcombined flat"}  
 
bool   fl_over       {no,prompt="Subtract overscan level"} 
bool   fl_trim       {yes,prompt="Trim off overscan region"}
bool   fl_bias       {yes,prompt="Subtract bias image"}
bool   fl_dark       {no,prompt="Subtract (scaled) dark image"}
bool   fl_fixpix     {yes,prompt="Interpolate across chip gaps"}
bool   fl_vardq      {no,prompt="Create variance and data quality frames"}
string bias          {"",prompt="Bias image"}
string dark          {"",prompt="Dark image"}
string key_exptime   {"EXPTIME",prompt="Exposure time header keyword"}
string key_biassec   {"BIASSEC",prompt="Header keyword for overscan strip image section."} 
string key_datasec   {"DATASEC",prompt="Header keyword for data section (excludes the overscan)"}
string rawpath       {"",prompt="GPREPARE: Path for input raw images"}
string sci_ext       {"SCI",prompt="Name of science extension"}
string var_ext       {"VAR",prompt="Name of variance extension"}
string dq_ext        {"DQ",prompt="Name of data quality extension"}
string key_mdf       {"MASKNAME",prompt="Header keyword for the MDF"}
string mdffile       {"",prompt="MDF to use if keyword not found"}
string mdfdir        {"gmos$data/", prompt="MDF database directory"}
string bpm           {"",prompt="Name of bad pixel mask file or image"}
string gaindb        {"gmos$data/gmosamps.dat",prompt="Database with gain data"}
string gratingdb     {"gmos$data/GMOSgratings.dat",prompt="Gratings database file"}
string filterdb      {"gmos$data/GMOSfilters.dat",prompt="Filters database file"}
string bpmfile       {"gmos$data/chipgaps.dat",prompt="Info on location of chip gaps"}
real   sat           {65000.,prompt="Saturation level in raw images"}
real   xoffset       {INDEF,prompt="X offset in wavelength [nm]"}
real   yoffset       {INDEF,prompt="Y offset in unbinned pixels"}

#Response/Profile fitting 
bool   fl_inter      {no,prompt="Fit response interactively?"}
bool   fl_answer     {yes,prompt="Continue interactive fitting?"}
string function      {"spline3",min="spline3|legendre|chebyshev|spline1",
                      prompt="Fitting function for response"}
int    order         {15,min=1,prompt="Order of fitting function"}
real   low_reject    {3,prompt="Low rejection in sigma of response fit"}
real   high_reject   {3,prompt="High rejection in sigma of response fit"}
int    niterate      {2,prompt="Number of rejection iterations in response fit"}

#Imcombine 
string combine       {"average",prompt="Combination operation"}
string reject        {"avsigclip",prompt="Rejection algorithm"}
string scale         {"mean",prompt="Image scaling"}
string zero          {"none",prompt="Image zeropoint offset"}
string weight        {"none",prompt="Image weights"}
string statsec       {"",prompt="Statistics section"}
real   lthreshold    {INDEF,prompt="Lower threshold"}
real   hthreshold    {INDEF,prompt="Upper threshold"}
int    nlow          {1,min=0,prompt="minmax: Number of low pixels to reject"}
int    nhigh         {1,min=0,prompt="minmax: Number of high pixels to reject"}
int    nkeep         {0,min=0,prompt="Minimum to keep or maximum to reject"}  
bool   mclip         {yes,prompt="Use median in sigma clipping algorithms?"}
real   lsigma        {3.,prompt="Lower sigma clipping factor"}
real   hsigma        {3.,prompt="Upper sigma clipping factor"}
string key_ron       {"RDNOISE",prompt="Keyword for readout noise in e-"}
string key_gain      {"GAIN",prompt="Keyword for gain in electrons/ADU"}
real   ron           {3.5,min=0.,prompt="Readout noise rms in electrons"}
real   gain          {2.2,min=0.00001,prompt="Gain in e-/ADU"}
string snoise        {"0.0",prompt="ccdclip: Sensitivity noise (electrons)"}
real   sigscale      {0.1,prompt="Tolerance for sigma clipping scaling correction"}
real   pclip         {-0.5,prompt="pclip: Percentile clipping parameter"}
real   grow          {0.0,prompt="Radius (pixels) for neighbor rejection"}

#Colbias
bool   ovs_flinter   {no,prompt="Interactive overscan fitting?"}
bool   ovs_med       {no,prompt="Use median instead of average in column bias?"} 
string ovs_func      {"chebyshev",min="spline3|legendre|chebyshev|spline1",
                       prompt="Overscan fitting function"}
int    ovs_order     {1,prompt="Order of overscan fitting function"}
real   ovs_lowr      {3.,prompt="Low sigma rejection factor"}
real   ovs_highr     {3.,prompt="High sigma rejection factor"}
int    ovs_niter     {2,prompt="Number of rejection iterations"}

#General 
string logfile       {"",prompt="Logfile name"}
bool   verbose       {yes,prompt="Verbose output?"}
int    status        {0,prompt="Exit status (0=good)"}
struct *scanfile     {"",prompt="Internal use only"} 


begin

#local variables of parameters 
string l_inflats, l_specflat, l_logfile, l_slitfunc
string l_bias, l_dark, l_bpm, l_key_biassec, l_key_datasec, l_ovs_func, l_combflat, l_bpmfile
string l_key_mdf, l_mdffile, l_combine, l_reject, l_scale, l_zero
string l_weight, l_statsec, l_key_exptime, l_key_ron, l_key_gain, l_snoise
string l_rfunction, l_sci_ext, l_var_ext, l_dq_ext
string l_mdfdir, l_gaindb, l_rawpath, l_gratingdb, l_filterdb
real   l_lthreshold, l_hthreshold, l_lsigma, l_hsigma,l_ron,l_gain,l_sigscale
real   l_pclip, l_grow, l_sat, l_ovs_lowr, l_ovs_highr, l_rlowrej, l_rhighrej
real   l_xoffset, l_yoffset
int    l_nlow, l_nhigh, l_nkeep, l_rorder, l_ovs_order, l_ovs_niter, l_niterate
bool   l_slitcorr, l_mclip, l_rinteractive, l_fl_vardq, l_verbose, l_redkeep
bool   l_ovs_flinter,l_ovs_med,l_fl_over,l_fl_bias,l_fl_dark,l_fl_trim, l_fl_fixpix
bool   l_fl_answer

bool   l_fl_gmosaic

#other parameters used within this task 

file temp1,temp2
bool reducing,gpflag,giflag,gsflag,gsfl1,mosflag,mosfl1
bool specmode 
string filt1,filtn,grat1,gratn,cutfile
string mdf,inlist,filelist,scilist,dqlist,img,specsec[500], imgraw
string response,varresponse,suf,combsig,combdq
string infile[20],dqout,tmpdq,headtest,nextens,next1
string scisec,varsec,sciflat,flatfit,varflat
string mscisec,mscifit,scorrsec,scorrvsec,ysec,inscoo
string gprep,gired,specsecfile,dqexp,tmplog
real firstexp,l_sgain,l_sron,gaineff,roneff,expdiff
int colpos1,colpos2,compos,seclen,x1,y1
int i,j,k,l,n,idum,nbad,len,nslits,nim
struct sdate

char l_sample
int  x11, x12, x21, x22, Xbin, Xmax


#Make parameter assignments  
l_inflats=inflats ; l_specflat=specflat
l_combflat=combflat ; l_redkeep=fl_keep
l_fl_over=fl_over ; l_fl_bias=fl_bias ; l_fl_dark=fl_dark ; l_fl_fixpix=fl_fixpix
l_bias=bias ; l_dark=dark ; l_bpm=bpm   ; l_bpmfile=bpmfile
l_key_mdf=key_mdf ; l_mdffile=mdffile 
l_key_biassec=key_biassec ; l_key_datasec=key_datasec 
l_ovs_flinter=ovs_flinter ; l_ovs_med=ovs_med 
l_ovs_func=ovs_func ; l_ovs_order=ovs_order  
l_slitcorr=fl_slitcorr ; l_slitfunc=slitfunc
l_combine=combine ; l_reject=reject
l_scale=scale ; l_zero=zero ; l_weight=weight
l_statsec=statsec ; l_key_exptime=key_exptime
l_lthreshold=lthreshold ; l_hthreshold=hthreshold
l_nlow=nlow ; l_nhigh=nhigh ; l_nkeep=nkeep
l_lsigma=lsigma ; l_hsigma=hsigma
l_verbose=verbose ; l_mclip=mclip
l_key_gain=key_gain ; l_key_ron=key_ron
l_gain=gain ; l_ron=ron ; l_snoise=snoise 
l_sigscale=sigscale ; l_pclip=pclip ; l_grow=grow
l_sci_ext = sci_ext ; l_dq_ext = dq_ext ; l_var_ext=var_ext
l_ovs_lowr=ovs_lowr ; l_ovs_highr=ovs_highr ; l_ovs_niter=ovs_niter
l_bpm=bpm ; l_fl_vardq=fl_vardq
l_rinteractive=fl_inter ; l_rorder=order ; l_rfunction=function
l_rlowrej=low_reject ; l_rhighrej=high_reject; l_niterate=niterate
l_logfile=logfile; l_sat=sat
l_fl_trim=fl_trim
l_mdfdir=mdfdir ; l_gaindb=gaindb ; l_rawpath=rawpath
l_gratingdb=gratingdb ; l_filterdb=filterdb
l_xoffset=xoffset ; l_yoffset=yoffset

status=0
l_fl_gmosaic=yes

# Keep some parameters from changing by outside world
cache("imgets","hedit","gprepare","gmos","tinfo","fparse","gimverify")

#Define temporary files
specsecfile=mktemp("tmpsecfile")
scilist=mktemp("tmpscilist")
tmplog=mktemp("tmplog")
if(!l_redkeep) { 
   l_combflat=mktemp("tmpcombflat")
}
if(l_fl_vardq){ 
dqlist=mktemp("tmpdqlist")
combsig=mktemp("tmpsig")
combdq=mktemp("tmpcombdq")
dqout=mktemp("tmpdqout")
tmpdq=mktemp("tmpdq") 
varsec=mktemp("tmpvarsec")
varresponse=mktemp("tmpvresponse")
varflat=mktemp("tmpvflat")
}
scisec=mktemp("tmpscisec")
response=mktemp("tmpresponse")
flatfit=mktemp("tmpfit")
sciflat=mktemp("tmpflat")
if(l_slitcorr) { 
   mscisec=mktemp("tmpmedflat")
   mscifit=mktemp("tmpmfitflat")
   scorrsec=mktemp("tmpscorr")
}
temp1=mktemp("tmpfilelist")
temp2=mktemp("tmpfilelist")
mdf=mktemp("tmpmdffile") 
mdf=mdf//".fits"

#Check logfile ... 
if (l_logfile == "STDOUT") {
    l_logfile = ""
    l_verbose=yes
} else if (l_logfile == "" || stridx(" ",l_logfile)>0) {
    l_logfile = gmos.logfile
    if (l_logfile == "" || stridx(" ",l_logfile)>0) {
        l_logfile = "gmos.log"
        printlog("WARNING - GSFLAT: both gsflat.logfile and gmos.logfile are empty.",l_logfile,l_verbose)
        printlog("                  Using default file gmos.log.",l_logfile,l_verbose)
    }
}
date | scan(sdate)
printlog("",l_logfile,l_verbose)
printlog("--------------------------------------------------------------------------------",l_logfile,l_verbose)
printlog("GSFLAT -- "//sdate,l_logfile,l_verbose)
printlog("",l_logfile,l_verbose)
printlog("Input images or list              = "//l_inflats,l_logfile,l_verbose)
printlog("Output spectral flat              = "//l_specflat,l_logfile,l_verbose)
printlog("Correct for slit function         = "//l_slitcorr,l_logfile,l_verbose) 
if(l_slitcorr) 
printlog("Slitfunc image from GSSLITFUNCTION= "//l_slitfunc,l_logfile,l_verbose)
printlog("Keep combined flat combflat       = "//l_redkeep,l_logfile,l_verbose) 
if(l_redkeep)
printlog("Combined flat (not normalized)    = "//l_combflat,l_logfile,l_verbose)
printlog("",l_logfile,l_verbose)
printlog("Fitting parameters for Spectral Flatfield: ",l_logfile,l_verbose)
printlog("  interactive = "//l_rinteractive,l_logfile,l_verbose)
printlog("  function    = "//l_rfunction,l_logfile,l_verbose)
printlog("  order       = "//l_rorder,l_logfile,l_verbose)
printlog("  low_reject  = "//l_rlowrej,l_logfile,l_verbose)
printlog("  high_reject = "//l_rhighrej,l_logfile,l_verbose)
printlog("  niterate    = "//l_niterate,l_logfile,l_verbose)
printlog("",l_logfile,l_verbose)
printlog("Overscan Subtraction fl_over   = "//l_fl_over,l_logfile,l_verbose)
printlog("Trim image           fl_trim   = "//l_fl_trim,l_logfile,l_verbose)
printlog("Bias Subtraction     fl_bias   = "//l_fl_bias,l_logfile,l_verbose)
printlog("Dark Subtraction     fl_dark   = "//l_fl_dark,l_logfile,l_verbose)
printlog("VAR & DQ planes      fl_vardq  = "//l_fl_vardq,l_logfile,l_verbose) 
printlog("Fixpix chip gaps     fl_fixpix = "//l_fl_fixpix,l_logfile,l_verbose
printlog("",l_logfile,l_verbose)
printlog("bias     = "//l_bias,l_logfile,l_verbose)
printlog("dark     = "//l_dark,l_logfile,l_verbose)
printlog("sci_ext  = "//l_sci_ext,l_logfile,l_verbose)
printlog("var_ext  = "//l_var_ext,l_logfile,l_verbose)
printlog("dq_ext   = "//l_dq_ext,l_logfile,l_verbose)
printlog("bpm      = "//l_bpm,l_logfile,l_verbose)
printlog("key_mdf  = "//l_key_mdf,l_logfile,l_verbose)
printlog("mdffile  = "//l_mdffile,l_logfile,l_verbose)
printlog("mdfdir   = "//l_mdfdir,l_logfile,l_verbose)
printlog("Chip gaps= "//l_bpmfile,l_logfile,l_verbose)
printlog("",l_logfile,l_verbose)

nbad=0

#The checks on things begin ... 

if (l_inflats == "" || l_inflats==" ") {
  printlog("ERROR - GSFLAT: Input Flats not specified",l_logfile,yes)
  nbad=nbad+1
}

# check existence of list files
if (substr(l_inflats,1,1) == "@") {
    inlist=substr(l_inflats,2,strlen(l_inflats))
    if(!access(inlist)) {
       printlog("ERROR - GSFLAT: Input list "//inlist//" not found",l_logfile,verbose+)
       nbad=nbad+1 
   }
}

if (l_specflat == "" || l_specflat==" ") {
  printlog("ERROR - GSFLAT: output specflat not specified",l_logfile,yes)
  nbad=nbad+1
}
if (l_redkeep && (l_combflat == "" || l_combflat==" ")) {
  printlog("ERROR - GSFLAT: output combflat not specified",l_logfile,yes)
  nbad=nbad+1
}

# Check that the output file does not already exist. If so, exit.
gimverify(l_specflat) ; l_specflat=gimverify.outname
if (gimverify.status != 1) {  
  printlog("ERROR - GSFLAT: Output file "//l_specflat//" already exists.",l_logfile,yes)
  nbad=nbad+1
}

# Check that the output combined flatfield does not already exist. If so, exit.
gimverify(l_combflat) ; l_combflat=gimverify.outname
if (gimverify.status  != 1) {  
  printlog("ERROR - GSFLAT: Output file "//l_combflat//" already exists.",l_logfile,yes)
  nbad=nbad+1
}

if(l_sci_ext =="" || l_sci_ext ==" ") {
  printlog("ERROR - GSFLAT: extension name sci_ext is missing",l_logfile,yes)
  nbad=nbad+1
}

#If dq/var propogation is requested, make sure the names are given 
if(l_fl_vardq) { 
   if (l_dq_ext=="" || l_dq_ext ==" ") {
       printlog("ERROR - GSFLAT: extension name dq_ext is missing",l_logfile,yes)
       nbad=nbad+1
   } else if (l_var_ext=="" || l_var_ext ==" ") { 
       printlog("ERROR - GSFLAT: extension name var_ext is missing",l_logfile,yes)
       nbad=nbad+1
   } 
}

#check no commas in sci_ext, var_ext and dq_ext
if(stridx(",",l_sci_ext)>0 || stridx(",",l_var_ext)>0 || stridx(",",l_dq_ext)>0 ) {
   printlog("ERROR - GSFLAT: sci_ext, var_ext or dq_ext contains commas, give root name only",l_logfile,yes)
   nbad=nbad+1
}

#Now check for slit-function file if slitcorr=yes is specified in the parameters
if (l_slitcorr) {
   gimverify(l_slitfunc) ; l_slitfunc=gimverify.outname
   if (gimverify.status != 0) { 
       printlog("ERROR - GSFLAT:  Input slit-function file : "//l_slitfunc//" either does not exist",l_logfile,yes)
       printlog("                 or is not a valid MEF.",l_logfile,yes)
       nbad=nbad+1
   } else { 
     imgets(l_slitfunc//"[0]","GSSLITFU") 
     if(imgets.value == "0" || imgets.value == " ") {
        printlog("ERROR - GSFLAT: Input file "//l_slitfunc//" was not produced by GSSLITFUNCTION.",l_logfile,yes)
        nbad=nbad+1
     }
   }
}

#check on existence of grating and filter databases, and chipgaps
if(!access(l_gratingdb)) { 
   printlog("ERROR - GSFLAT: gratings database file : "//l_gratingdb//" does not exist.",l_logfile,yes)
   nbad=nbad+1
}
if(!access(l_filterdb)) { 
   printlog("ERROR - GSFLAT: filters database file : "//l_filterdb//" does not exist.",l_logfile,yes)
   nbad=nbad+1
}
if(l_gaindb != "" && !access(l_gaindb)) { 
   printlog("ERROR - GSFLAT: gain database file : "//l_gaindb//" does not exist.",l_logfile,yes)
   nbad=nbad+1
}
if(l_bpmfile != "" && !access(l_bpmfile)) { 
   printlog("ERROR - GSFLAT: Chip gap definition file : "//l_bpmfile//" does not exist.",l_logfile,yes)
   nbad=nbad+1
} else if (l_bpmfile != "") {
  fields(l_bpmfile,"1,2",lines="2") | scan(x11,x12)
  fields(l_bpmfile,"1,2",lines="3") | scan(x21,x22)
  l_sample="1:"//str(x11-1)//","//str(x12+1)//":"//str(x21-1)//","//str(x22+1)//":6218"
} else
  l_sample="*"

#Assuming we're ok so far, check that all images in the list exist, and are of the correct filetype
#

if(nbad == 0) { 

#Create list of input flat images 
files(l_inflats, > temp1)
scanfile=temp1 

nim=0
while(fscan(scanfile,img) != EOF) {
      nim=nim+1 
      gimverify(img) ; img=gimverify.outname//".fits" ; imgraw=gimverify.outname//".fits"
      if(gimverify.status>=1 && l_rawpath!="") {
          gimverify(l_rawpath//img) ; imgraw=gimverify.outname//".fits"
      }
      if(gimverify.status != 0) { 
        if(gimverify.status == 1) { 
           printlog("ERROR - GSFLAT: Input image : "//img//" does not exist.",l_logfile,yes)
           nbad=nbad+1 
        } else if (gimverify.status == 2) { 
           printlog("ERROR - GSFLAT: Input image : "//img//" is in OIF format, not MEF.",l_logfile,yes)
           nbad=nbad+1
        } else if (gimverify.status == 3) { 
           printlog("ERROR - GSFLAT: Input image : "//img//" is a GEIS file, not MEF.",l_logfile,yes)
           nbad=nbad+1
        } else if (gimverify.status == 4) { 
           printlog("ERROR - GSFLAT: Input image : "//img//" is a simple FITS file, not MEF.",l_logfile,yes)
           nbad=nbad+1
        }
      }

}
# Exit if problems found
if (nbad > 0) {
    printlog("ERROR - GSFLAT: "//nbad//" fatal errors found.",l_logfile,yes)
    goto error
}

scanfile=temp1 

nim=0
while(fscan(scanfile,img) != EOF) {
      nim=nim+1 
      gimverify(img) ; img=gimverify.outname//".fits" ; imgraw=gimverify.outname//".fits"
      if(gimverify.status>=1 && l_rawpath!="") {
          gimverify(l_rawpath//img) ; imgraw=gimverify.outname//".fits"
      }

# Check for existing GSREDUCEd image 
     if(imaccess("gs"//img)) {
       imgets("gs"//img//"[0]","GSREDUCE")
       if(imgets.value!="" && imgets.value!=" " && imgets.value!="0") {
         printlog("GSFLAT: Using image gs"//img//" previously processed with GSREDUCE",l_logfile,l_verbose)
         img="gs"//img
         infile[nim]=img
       } else {
         printlog("ERROR - GSFLAT: Found output image gs"//img//" which is not GSREDUCEd",l_logfile,yes)
         nbad=nbad+1
       }
     } else {
       infile[nim]=img
       img=imgraw
     }
     
      imgets(img//"[0]","NEXTEND", >>& "dev$null") 
      nextens=int(imgets.value) 

      if(nim == 1) {
         next1=nextens
      }
      if(nextens != next1) {
      printlog("ERROR - GSFLAT: Input images do not all have the same number of extensions",
         l_logfile,yes)
         nbad=nbad+1
      } 

      imgets(img//"[0]","MASKTYP", >>& "dev$null") 
      if(imgets.value == "1" ) {
         specmode=yes
      } else { 
         specmode=no
      }
      if(!specmode) { 
           printlog("ERROR - GSFLAT: Input file "//img//" not of type LONGSLIT or MOS.",l_logfile,yes)
           nbad=nbad+1 
      }

# -------------
#check if GSREDUCE has been run if reductions are not requested here
      imgets(img//"[0]","GSREDUCE", >>& "dev$null")
      if(imgets.value == " " || imgets.value == "" || imgets.value == "0") {
         gsflag=no
         if(nim==1)
           gsfl1=no
         else if (gsflag!=gsfl1) {
             printlog("ERROR - GSFLAT: Cannot mix raw/GPREPARE'd images with GSREDUCE'd images",
             l_logfile,yes)
             nbad=nbad+1
         }
      } else  {
         gsflag=yes
         if(nim==1)
           gsfl1=yes
         else if (gsflag!=gsfl1) {
             printlog("ERROR - GSFLAT: Cannot mix raw/GPREPARE'd images with GSREDUCE'd images",
             l_logfile,yes)
             nbad=nbad+1
         }
      }

      if(!gsflag && (!l_fl_over && !l_fl_bias && !l_fl_dark &&!l_fl_trim)){
          printlog("ERROR - GSFLAT: Flat Image - "//img//" has not been GSREDUCED",l_logfile,yes)
          printlog("                and no reduction flags are requested here.",l_logfile,yes)
          nbad=nbad+1
#          imgets(img//"[0]","GPREPARE", >& "dev$null")
#          if(imgets.value != "" && imgets.value != " ") {
#             printlog("ERROR - GSFLAT: Image is GPREPARED, however no reduction flags are set.",l_logfile,yes)
#             printlog("                You need to process with GSREDUCE or specify those options here.",l_logfile,yes)
#             nbad=nbad+1
#          }
      }

      if(gsflag) {
         if(l_fl_over || l_fl_bias || l_fl_dark || l_fl_trim) { 
           printlog("WARNING - GSFLAT: Image "//img//" has been processed with GSREDUCE",
             l_logfile,yes)
           printlog("                  Ignoring the processing flags",l_logfile,yes)
           l_fl_over=no ; l_fl_bias=no ; l_fl_dark=no ; l_fl_trim=no
         }
      }

      imgets(img//"[0]","GMOSAIC", >& "dev$null")
      if(imgets.value != " " && imgets.value != "" && imgets.value!="0") {
         printlog("WARNING - GSFLAT: Image "//img//" has been processed with GMOSAIC",
            l_logfile,yes)
         mosflag=yes
         if(l_fl_over || l_fl_bias || l_fl_dark || l_fl_trim) { 
           printlog("                  Ignoring the processing flags",l_logfile,yes)
           l_fl_over=no ; l_fl_bias=no ; l_fl_dark=no ; l_fl_trim=no
         } else {
           printlog("                  No mosaicing done",l_logfile,yes)
           l_fl_gmosaic=no
         }
         if(nim==1)
           mosfl1=yes
         else if (mosflag!=mosfl1) {
           printlog("ERROR - GSFLAT: Cannot mix raw/GPREPARE'd images with GMOSAIC'd images",
                l_logfile,yes)
           nbad=nbad+1
         }
      } else {
        mosflag=no
         if(nim==1)
           mosfl1=no
         else if (mosflag!=mosfl1) {
           printlog("ERROR - GSFLAT: Cannot mix raw/GPREPARE'd images with GMOSAIC'd images",
                l_logfile,yes)
           nbad=nbad+1
         }
      }


# -------------
#check exposure times
      imgets(img//"[0]",l_key_exptime,>& "dev$null") 
      if (imgets.value == "0" || imgets.value == " "){
           printlog("ERROR - GSFLAT: Image header parameter "//l_key_exptime//" not found",l_logfile,yes)
           nbad=nbad+1
      }
      if (nim == 1) {
          firstexp=real(imgets.value)
      } else {
        expdiff=(real(imgets.value)-firstexp)/firstexp
        if (expdiff > 0.1){
            printlog("WARNING - GSFLAT: "//img//" - exposure time different by more than 10%",l_logfile,yes)
        }
      }


#Check that images in list all have the same FILTER1 and GRATING ID's and the same central wavelength. 
      imgets(img//"[0]","FILTER2")
      if (nim == 1) { 
          filt1=(imgets.value)
      }
      filtn=(imgets.value)
      imgets(img//"[0]","GRATING")
      if (nim == 1) { 
          grat1=(imgets.value)
      }
      gratn=(imgets.value)
      if(filtn != filt1){
         printlog("ERROR - GSFLAT: Input images do not all have the same filter ID.",l_logfile,yes)
         nbad=nbad+1
      }
      if(gratn != grat1){
         printlog("ERROR - GSFLAT: Input images do not all have the same grating ID.",l_logfile,yes)
         nbad=nbad+1
      }

#End of while loop over images
}
scanfile=""
#End of if-loop for "nbad=0 so far" 
}


# Exit if problems found
if (nbad > 0) {
    printlog("ERROR - GSFLAT: "//nbad//" fatal errors found.",l_logfile,yes)
    goto error
}

#END OF BASIC CHECKS ...
#Make sure the input names has .fits on them
delete(temp1,verif-, > "dev$null")
for(i=1;i<=nim;i+=1){
    files(infile[i], >> temp1)
}
scanfile=temp1

#NOW, call GSREDUCE to overscan, bias-subtract, and dark-subtract if requested (most likely)
reducing=no
if(l_fl_over || l_fl_bias || l_fl_dark || l_fl_trim) {
   reducing=yes
   printlog(" ",l_logfile,l_verbose)
   printlog("GSFLAT: Calling GSREDUCE.",l_logfile,l_verbose)
   printlog(" ",l_logfile,l_verbose)
   gsreduce("@"//temp1,outimages="",outpref="gs",logfile=l_logfile,
    verbose=l_verbose,fl_gmosaic=l_fl_gmosaic,fl_fixpix=l_fl_fixpix,fl_over=l_fl_over,
    fl_bias=l_fl_bias,fl_dark=l_fl_dark,fl_flat-,fl_cut-,bias=l_bias,dark=l_dark,flatim="",
    fl_title=no,bpmfile=l_bpmfile,
    key_exptime=l_key_exptime,key_biassec=l_key_biassec,
    key_datasec=l_key_datasec,fl_vardq=l_fl_vardq,sci_ext=l_sci_ext,
    var_ext=l_var_ext,dq_ext=l_dq_ext,key_mdf=l_key_mdf,mdffile=l_mdffile,
    mdfdir=l_mdfdir,bpm=l_bpm,key_ron=l_key_ron,key_gain=l_key_gain,
    ron=l_ron,gain=l_gain,sat=l_sat,ovs_flinter=l_ovs_flinter,
    ovs_med=l_ovs_med,ovs_func=l_ovs_func,ovs_order=l_ovs_order,
    ovs_lowr=l_ovs_lowr,ovs_highr=l_ovs_highr,ovs_niter=l_ovs_niter,
    fl_trim=l_fl_trim,gaindb=l_gaindb,rawpath=l_rawpath,
    gratingdb=l_gratingdb,filterdb=l_filterdb)

   if(gsreduce.status != 0) { 
      printlog("ERROR - there was a problem running GSREDUCE.  Stopping now.",l_logfile,yes)
      nbad=nbad+1
      goto error
   } 

   printlog("GSFLAT: Returned from GSREDUCE.",l_logfile,l_verbose)
   for(i=1;i<=nim;i+=1){
       files("gs"//infile[i], >> temp2)
   }
   scanfile=temp2
}

#Start Process of combining flats 
#i loop is over the # of images

i=0
while(fscan(scanfile,img) !=EOF) { 

i=i+1

printlog("GSFLAT: Image #"//i//" - Working on : "//img,l_logfile,l_verbose)

#check gain and readnoise values - input parameters vs. header keywords/values. 
    if (l_key_gain != "" || l_key_gain != " ") {
        imgets(img//"[0]",l_key_gain, >& "dev$null")
        if (imgets.value == "0") {
            printlog("WARNING - GSFLAT: keyword "//l_key_gain//" not found in "//img,l_logfile,l_verbose)
            printlog("                  Using GAIN = "//str(l_gain),l_logfile,yes)
            l_sgain=l_gain
        } else {
          l_sgain=real(imgets.value)
        }
    } else {
      l_sgain=l_gain
    }

    if (l_key_ron != "" || l_key_ron != " ") {
        imgets(img//"[0]",l_key_ron, >& "dev$null")
        if (imgets.value == "0") {
            printlog("WARNING - GSFLAT: keyword "//l_key_ron//" not found in "//infile[i],l_logfile,l_verbose)
            printlog("                  Using RON = "//str(l_ron),l_logfile,yes)
            l_sron=l_ron
        } else {
          l_sron=real(imgets.value)
        }
    } else {
      l_sron=l_ron
    }

#copy MDF file to disk for later use 
# Check that the MDF exists
tinfo(img//"[MDF]",ttout-, >& "dev$null")
if (tinfo.tbltype!="fits") {
    printlog("ERROR - GSFLAT: MDF file does not exist",l_logfile,yes)
    nbad=nbad+1
    goto error
}
tcopy(img//"[MDF]",mdf,verbose=no, >& "dev$null")

#Make list for imcombine of science extensions 
print(img//"["//l_sci_ext//"]", >> scilist)

#Copy DQ extensions
if (l_fl_vardq) { 
    imcopy(img//"["//l_dq_ext//"]",tmpdq//"_"//str(i)//".pl",verbose-)
    print(tmpdq//"_"//str(i)//".pl", >> dqlist)
    hedit(img//"["//l_sci_ext//"]","BPM",tmpdq//"_"//str(i)//".pl",add+,update+,verify-,show-)
}

#end of while loop 
}

#Now, combine images
#special cases for low numbers of images

printlog(" ",l_logfile,l_verbose)

if (nim == 1) {
    printlog("WARNING - GSFLAT: only one image.",l_logfile,yes)
    imcopy(img//"["//l_sci_ext//"]",l_combflat,verbose-)
    if(l_fl_vardq) { 
       imcopy(tmpdq//"_1.pl",combdq,verbose-)
       imcopy(infile[1]//"["//l_var_ext//"]",combsig,verbose-)
    }
} else {
  if (nim <= 5){
      if (nim == 2){
          printlog("WARNING - GSFLAT: only combining two images, turning off rejection",l_logfile,yes)
          printlog("                  and setting combine type to average.",l_logfile,yes)
          l_reject="none"
          l_combine="average" 
      } else {
        printlog("GSFLAT: combining five or less images.",l_logfile,yes)
      }
  }

#Note: output of imcombine will be a FITS file, not a MEF 
   
if(l_fl_vardq){          
   imcombine("@"//scilist,l_combflat,rejmask="",plfile=combdq,sigma=combsig,\
             combine=l_combine,reject=l_reject,project=no,outtype="real",\
             offsets="none",masktype="goodvalue",maskvalue=0,blank=0,\
             scale=l_scale,zero=l_zero,weight=l_weight,lsigma=l_lsigma,\
             hsigma=l_hsigma,\
             statsec=l_statsec,expname=l_key_exptime,lthreshold=l_lthreshold,\
             hthreshold=l_hthreshold,nlow=l_nlow,nhigh=l_nhigh,nkeep=l_nkeep,\
             mclip=l_mclip,rdnoise=l_sron,gain=l_sgain,snoise=l_snoise,\
             sigscale=l_sigscale,pclip=l_pclip,grow=l_grow,logfile=tmplog)

   if (access(tmplog)) {
       if (l_logfile != "" && l_logfile != " ") {
           type(tmplog, >> l_logfile)
       }
       if (l_verbose) {
           type(tmplog)
       }
       delete(tmplog,verify-, >& "dev$null") 
   }
 
   # make variance image by squaring combsig
   imarith(combsig,"*",combsig,combsig,verbose-)


   # Calculate the DQ plane (this renormalizes it to 0:1 rather than the # of images
   dqexp = "((a == b) ?1 : 0)"
   imexpr(dqexp,dqout,combdq,nim,dims="auto",intype="auto",outtype="int",refim="auto",bwidth=0,
          btype="nearest",bpixval=0.,rangecheck+,verbose-,exprdb="none",lastout="")

   delete(dqlist,verify-, >& "dev$null")
   # clean headers
   hedit("@"//scilist,"BPM","",delete+,update+,verify-,show-)
   imdelete(tmpdq//"*.pl",verify-, >>& "dev$null")

#if don't want DQ/VAR propogation ... 
} else if(!l_fl_vardq){ 
   
  imcombine("@"//scilist,l_combflat,rejmask="",plfile="",sigma="",
            combine=l_combine,reject=l_reject,project=no,outtype="real",
            offsets="none",masktype="goodvalue",maskvalue=0,blank=0,
            scale=l_scale,zero=l_zero,weight=l_weight,lsigma=l_lsigma,hsigma=l_hsigma,
            statsec=l_statsec,expname=l_key_exptime,lthreshold=l_lthreshold,
            hthreshold=l_hthreshold,nlow=l_nlow,nhigh=l_nhigh,nkeep=l_nkeep,
            mclip=l_mclip,rdnoise=l_sron,gain=l_sgain,snoise=l_snoise,
            sigscale=l_sigscale,pclip=l_pclip,grow=l_grow,logfile=tmplog)

  if (access(tmplog)) {
      if (l_logfile != "" && l_logfile != " ") {
          type(tmplog, >> l_logfile)
      }
      if (l_verbose) {
          type(tmplog)
      }
      delete(tmplog,verify-, >& "dev$null") 
  }

}

  printlog("GSFLAT: imcombine complete.  Output in file : "//l_combflat,l_logfile,l_verbose)
}  # End of more than one image

delete(scilist,verify-, >> "dev$null")

hedit(l_combflat,"BPM",value="",add-,delete+,updat+,verif-,show-, >& "dev$null")
printlog(" ",l_logfile,l_verbose)

#VAR plane is now in combsig and DQ in dqout (if requested). The combined flat is l_combflat 
#For LONGSLIT mode the MDF will have 3 lines (2 bridges, three spectra sections) We want to keep
#these in the same SCI extensions however even though we will do some of the calibrations separately 
#for each section.  GSCUT will now do this rather than this task, but some things we still need to take
#care of here.

tinfo(mdf,tbltype="fits",subtype="binary",ttout=no)
nslits=tinfo.nrows
printlog("GSFLAT : Found "//str(nslits)//" Slits in MDF.",l_logfile,l_verbose) 

# make dummy images for inserting response sections for both science and variance extensions (if needed)
#
   printlog(" ",l_logfile,l_verbose)
   imarith(l_combflat,"/",l_combflat,response,verbose-)
   if(l_fl_vardq) {
      imarith(combsig,"*",0.,varresponse,verbose-)
   }


#Call gscut to get the images sections for each spectrum in the MDF 

if(reducing) {
  cutfile="gs"//infile[1]
} else { 
  cutfile=infile[1]
}
printlog("GSFLAT: Calling GSCUT to determine image sections for spectra.",l_logfile,l_verbose)
gscut(cutfile,outimag="",secfile=specsecfile,logfile=l_logfile,
      fl_vard=l_fl_vardq,gratingdb=l_gratingdb,filterdb=l_filterdb,
      verbose=l_verbose,xoffset=l_xoffset,yoffset=l_yoffset) 

if(!access(specsecfile) ) { 
   printlog("ERROR - GSFLAT: GSCUT failed in some way. ",l_logfile,yes)
   nbad=nbad+1
   goto error
} 
printlog("GSFLAT: Returned from GSCUT.",l_logfile,l_verbose)

#Now we process each slitlet 

count(specsecfile) | scan(nslits)

scanfile=specsecfile
i=1

# Set the sample to avoid the chip gaps, if possible
if(l_sample!="*") {
  imgets(l_combflat,"CCDSUM", >>& "dev$null")
  print(imgets.value) | scan(Xbin)
  imgets(l_combflat,"i_naxis1", >>& "dev$null")
  Xmax=int(imgets.value) 
  x11=int((x11-1)/Xbin+0.5) ; x12=int((x12-1)/Xbin+0.5)
  x21=int((x21-1)/Xbin+0.5) ; x22=int((x22-1)/Xbin+0.5)
  l_sample="1:"//str(x11-1)//","//str(x12+1)//":"//str(x21-1)//","//str(x22+1)//":"//Xmax
}
printlog(" ",l_logfile,l_verbose)
printlog("GSFLAT: Begin fitting response functions for slit(s)",l_logfile,l_verbose)
printlog("GSFLAT: Sample in fitting "//l_sample,l_logfile,l_verbose)
printlog(" ",l_logfile,l_verbose)

while(fscan(scanfile,s1) !=EOF) { 

     inscoo=mktemp("tmpinscoo")
     specsec[i]=s1
     #unfortunately, while the specsec format is great for cutting out image pieces, it doesn't
     #work for iminsert later on which needs the (x,y) of the bottom-left-corner. So, we'll have
     #to rip apart the string to find these values...
     colpos1=stridx(":",s1)
     compos=stridx(",",s1)
     seclen=strlen(s1)
     x1=int(substr(s1,2,colpos1-1))
     ysec=substr(s1,compos+1,seclen)
     colpos2=stridx(":",ysec)
     y1=int(substr(ysec,1,colpos2-1))
     print(x1," ",y1, > inscoo) 
     
     if(l_verbose){
       printlog("GSFLAT: Fitting response for slit#"//str(i)//" - SpecSec = "//specsec[i],l_logfile,l_verbose)
     }
     
     imcopy(l_combflat//specsec[i],scisec,verbose-)

     if(l_fl_vardq) {
        imcopy(combsig//specsec[i],varsec,verbose-)
     }
     if(l_slitcorr) { 
        imcopy(l_slitfunc//"["//l_sci_ext//"]"//specsec[i],scorrsec,verbose-)
        imcopy(l_slitfunc//"["//l_var_ext//"]"//specsec[i],scorrvsec,verbose-)
     }

#if using a slit function image from gsslitfunction then we only want to normalize the lamp flat but leave 
#the profile, otherwise we will use fit1d directly which effectively removes the slitfunction from the data.      
   if(!l_slitcorr) {
      fit1d(scisec,flatfit,type="fit",axis=1,interactive=l_rinteractive,
       sample=l_sample,naverage=1,function=l_rfunction,order=l_rorder,
       low_reject=l_rlowrej,high_reject=l_rhighrej,niterate=l_niterate,grow=0.)
      imarith(scisec,"/",flatfit,sciflat,verbose-)
      if(l_fl_vardq) { 
      imarith(flatfit,"*",flatfit,flatfit,verbose-)
      imarith(varsec,"/",flatfit,varflat,verbose-)
      }
   #Now, if we have an illumination correction from gsslitfunction, we want to leave the profile in the 
   #flatfield and take it out with the slitfunction (this will all be much slower) 
    } else if (l_slitcorr) {
      #First, median (maybe could boxcar?) perpendicular to dispersion so that we don't remove the slit
      #profile in the flat
      median(scisec,mscisec,xwin=1,ywin=specwid,zloreject=3.,zhireject=3.,
        boundary="nearest",constant=0.,verbose=no)
      #fit1d(mscisec,mscifit,type="fit",axis=1,interac-,order=l_rorder,
        #low_reject=l_rlowrej,high_reject=l_rhighrej,niterate=l_niterate,grow=0)
      fit1d(mscisec,mscifit,type="fit",axis=1,interactive=l_rinteractive,
       sample=l_sample,naverage=1,function=l_rfunction,order=l_rorder,
       low_reject=l_rlowrej,high_reject=l_rhighrej,niterate=l_niterate,grow=0.)
      imarith(scisec,"/",mscifit,sciflat,verbose-)
      imarith(sciflat,"/",scorrsec,sciflat,verbose-)
      if(l_fl_vardq) { 
         imarith(mscifit,"*",mscifit,flatfit,verbose-)
         imarith(varsec,"/",flatfit,varflat,verbose-)
         imarith(varflat,"/",scorrvsec,varflat,verbose-)
      }
    }
    # try to get rid of red emission lines 
    #imreplace(sciflat,1.,lower=1.1,upper=INDEF,radius=3)
    # insert sections into larger images
    iminsert(response,sciflat,response,"replace",coordfile=inscoo,offset1=0,offset2=0,xcol="c1",ycol="c2")
    if(l_fl_vardq) { 
       iminsert(varresponse,varflat,varresponse,"replace",coordfile=inscoo,offset1=0,offset2=0,xcol="c1",ycol="c2")
       imdelete(varsec//","//varflat,verify-)
    }
    #cleaning up a bit 
    imdelete(scisec//","//sciflat//","//flatfit,verify-, >& "dev$null")
    if(l_slitcorr) { 
       imdelete(mscisec//","//mscifit//","//scorrsec,verify-)
       if(l_fl_vardq) { 
          imdelete(scorrvsec,verify-)
       }
    }
    i=i+1 
    delete(inscoo,verify-)

# Give the user a chance to fit the rest non-interactively
   if(l_rinteractive && nslits>1) {
      l_fl_answer=fl_answer
      l_rinteractive=l_fl_answer
   }

#end of while loop over slits 
}
scanfile=""
printlog(" ",l_logfile,l_verbose)
printlog("GSFLAT: Finished Loop over slits. Now packing output MEF...",l_logfile,l_verbose)

date | scan(sdate) 

# update headers
hedit(response,"BPM","",delete+,verify-,show-)
if(l_fl_vardq) {
   hedit(varresponse//","//dqout,"BPM","",delete+,verify-,show-)
}

#Copy the PHU from the first of the "reduced" flatfields 
suf = substr(l_specflat,strlen(l_specflat)-3,strlen(l_specflat))
if (suf!="fits") { 
    l_specflat=l_specflat//".fits"
}

wmef(input=mdf,output=l_specflat,extnames="MDF",phu=cutfile,verbose-, >& "dev$null")
if(wmef.status != 0) { 
   printlog("ERROR - GSFLAT: problem writing output MEF.",l_logfile,yes)
   nbad=nbad+1
   goto error 
} 

#Now the science extension/flatfield
fxinsert(response//".fits",l_specflat//"[1]",groups="",verbose=yes, >& "dev$null")
gemhedit(l_specflat//"[2]","EXTNAME",l_sci_ext,"Extension name")
gemhedit(l_specflat//"[2]","EXTVER",1,"Extension number")
#gemhedit(l_specflat//"[2]","GEM-TLM",sdate,"Time of last modification with Gemini") 
hedit(l_specflat//"[2]","BPM",value="",add-,delete+,ver-,show-)

#Now, the VAR/DQ planes if requested ...
if (l_fl_vardq) {
    hedit(varresponse//".fits","OBJECT",value="",upd+,ver-,show-)
    fxinsert(varresponse//".fits",l_specflat//"[2]",groups="",verbose=yes, >& "dev$null")
    fxinsert(dqout//".fits",l_specflat//"[3]",groups="",verbose=yes, >& "dev$null")
    gemhedit(l_specflat//"[3]","EXTNAME",l_var_ext,"Extension number")
    gemhedit(l_specflat//"[3]","EXTVER",1,"Extension number")
#    gemhedit(l_specflat//"[3]","GEM-TLM",sdate,"Time of last modification with Gemini")
    hedit(l_specflat//"[3]","BPM",value="",add-,delete+,ver-,show-)
    gemhedit(l_specflat//"[4]","EXTNAME",l_dq_ext,"Extension name")
    gemhedit(l_specflat//"[4]","EXTVER",1,"Extension number")
#    gemhedit(l_specflat//"[4]","GEM-TLM",sdate,"Time of last modification with Gemini")
    hedit(l_specflat//"[4]","BPM",value="",add-,delete+,ver-,show-)
    gemhedit(l_specflat//"[0]","NEXTEND",4,"Number of extensions")
} else { 
  gemhedit(l_specflat//"[0]","NEXTEND",2,"Number of extensions")
}

#Update the PHU
hedit(l_specflat//"[0]","BPM","",add-,del+,upd+,ver-,show-)
gemhedit(l_specflat//"[0]","NCOMBINE",nim,"Number of images in IMCOMBINE")
gemhedit(l_specflat//"[0]","GSFLAT",sdate,"Time stamp for GSFLAT")
gemhedit(l_specflat//"[0]","GEM-TLM",sdate,"Time of last modification with Gemini")
if(l_slitcorr) {
   gemhedit(l_specflat//"[0]","GSSLITFUNC",sdate,"Time stamp for GSSLITFUNCTION")
}

gemhedit(l_specflat//"[0]","GAINORIG",l_sgain,"Original Gain value")
gemhedit(l_specflat//"[0]","RONORIG",l_sron,"Original Read-noise value")
gaineff=real(nim)*l_sgain
roneff=l_sron/sqrt(real(nim))
if(l_key_gain == " " || l_key_gain == "") { 
   l_key_gain="GAIN"
} 
if (l_key_ron == " " || l_key_ron == "") { 
  l_key_ron="RON"
} 
gemhedit(l_specflat//"[0]",l_key_gain,gaineff,"Rescaled Gain value")
gemhedit(l_specflat//"[0]",l_key_ron,roneff,"Rescaled Read-noise value")



# clean up
goto clean

error:
  status=1
  print(" ")
  printlog("ERROR - GSFLAT: Program execution failed with "//str(nbad)//" errors.",l_logfile,verbose=yes) 
  goto clean

clean:
  if(nbad == 0) { 
     printlog(" ",l_logfile,yes) 
     printlog("GSFLAT done  "//sdate,l_logfile,yes)
     printlog(" ",l_logfile,yes)
  }
  delete("tmpmdf*,"//temp1//","//temp2//","//specsecfile,verif-, >>& "dev$null")
  imdelete(response,verif-, >>& "dev$null")
  if (l_fl_vardq) {
      imdelete(combsig//","//combdq//","//dqout//","//varresponse,verif-, >>& "dev$null")
  }
  if(!l_redkeep) { 
      imdelete(l_combflat,verif-, >>& "dev$null") 
  } 
  # close log file
  printlog("--------------------------------------------------------------------------------",l_logfile,l_verbose)
  scanfile=""

end
