# Copyright(c) 2001-2002 Association of Universities for Research in Astronomy, Inc.

procedure nisky(inimages)

# Make an average or median sky image for NIRI images, removing
# stars first.  Input images (inimages) are the images to combine.
# 
# Version  June 19, 2001  JJ, beta-release
#          Oct 15, 2001  JJ, v1.2 release
#          Nov 16, 2001  JJ, automatic output filename
#          Jan 3, 2002   JJ, fixed bug with fl_vardq+ but DQ doesn't exist
#                            and you want VAR and DQ in the output (as can 
#                            be the case with skyflats)
#          Jan 15, 2002  JJ, added checks for INDEF in object list; adjusted
#                            minimum radius a bit
#          Feb 28, 2002  JJ, v1.3 release
#          Mar 25, 2002  JJ, fixed bug in fixing nonlinear and saturated pix
#                            that kept daofind from finding bright stars
#          May 29, 2002  JJ, removed WMEF calls,modified scilist
#          May 30, 2002  JJ, headers written all at once, debug
#          Jun 07, 2002  JJ, delete output if exiting with ERROR
#                            added logic to avoid unnecessary nifastsky call
#                            fixed bug for repeated input file names
#          Jun 12, 2002  JJ  fixed bug when input file doesn't exist
#          Sept 10, 2002 IJ  parameter encoding
#          Sept 20, 2002 JJ  v1.4 release

char    inimages    {prompt="Raw NIRI image list to combine"}                   # OLDP-1
char    outimage    {"",prompt="Output sky image"}                              # OLDP-1
char    outtitle    {"default",prompt="Title for output image"}                 # OLDP-3
char    combtype    {"default",prompt="Type of combine operation",enum="default|median|average"} # OLDP-2
char    rejtype     {"avsigclip",prompt="Type of rejection",enum="none|avsigclip|minmax"} # OLDP-2
char    logfile     {"",prompt="Name of log file"}                              # OLDP-1
char    statsec     {"[100:924,100:924]",prompt="Statistics section"}           # OLDP-2
int     nlow        {0,min=0,prompt="minmax: Number of low pixels to reject"}   # OLDP-2
int     nhigh       {1,min=0,prompt="minmax: Number of high pixels to reject"}  # OLDP-2
real    lsigma      {3.,min=0.,prompt="avsigclip: Lower sigma clipping factor"} # OLDP-2
real    hsigma      {3.,min=0.,prompt="avsigclip: Upper sigma clipping factor"} # OLDP-2
real    threshold   {4.5,min=1.5,prompt="Threshold in sigma for object detection"} # OLDP-2
real    fwhmpsf     {4.5,min=0.5,prompt="Estimate of PSF FWHM in pixels"}     # OLDP-2
real    rebin       {7.0,min=4.0,prompt="Minimum PSF FWHM (pix) for rebinning"} # OLDP-3
char    key_exptime {"EXPTIME",prompt="Keyword for exposure time"}            # OLDP-3
char    key_ron     {"RDNOISE",prompt="Keyword for readout noise in e-"}      # OLDP-3
char    key_gain    {"GAIN",prompt="Keyword for gain in electrons/ADU"}       # OLDP-3
char    key_sat     {"SATURATI",prompt="Keyword for saturation level in ADU"} # OLDP-3
char    key_filter  {"FILTER",prompt="Header keyword for filter id"}          # OLDP-3
char    key_airmass {"AIRMASS",prompt="Header keyword for airmass"}           # OLDP-3
char    masksuffix  {"msk",prompt="Mask name suffix"}                         # OLDP-3
real    maskfactor  {1.,min=0.01,prompt="Scaling factor to change the size of mask holes"} # OLDP-2
bool    fl_keepmasks {no,prompt="Keep object masks for each input image?"}    # OLDP-2
bool    fl_gemseeing {no,prompt="Use gemseeing to refine the PSF FWHM?"}      # OLDP-2
char    sci_ext     {"SCI",prompt="Name or number of science extension"}      # OLDP-3
char    var_ext     {"VAR",prompt="Name or number of variance extension"}     # OLDP-3
char    dq_ext      {"DQ",prompt="Name or number of data quality extension"}  # OLDP-3
bool    fl_vardq    {yes,prompt="Create variance and data quality frames in output?"} # OLDP-3
bool    fl_dqprop   {no,prompt="Retain input DQ information in output?"}      # OLDP-3
bool    verbose     {yes,prompt="Verbose actions"}                            # OLDP-2
int     status      {0,prompt="Exit status (0=good)"}                         # OLDP-4
struct  *scanfile   {prompt="Internal use only"}                              # OLDP-4

begin

#-------------------------------------------------------------------------
# declare local variables
char    l_inimages, l_outimage, l_combtype, l_rejtype, l_logfile
char    l_key_ron, l_key_gain, l_key_filter, l_key_airmass, l_key_sat 
char    l_key_exptime, l_masksuffixreal, l_outtitle, l_statsec
real    l_fwhmpsf, l_threshold, l_datamax, l_lsigma, l_hsigma
bool    l_verbose, l_fl_gemseeing, l_fl_keepmasks, l_fl_objects
int     l_nlow, l_nhigh, l_xsize, l_ysize, nfiles
int     i, n, nbad, l_binfactor, l_nempty, l_nbadpix, x1, x2, y1, y2
real    skyconst, l_ron, l_gain, l_maskfactor
char    img, firstimage, imgsci, imgdq, mask
char    tmpfile1, tmpfile2, tmpimg, tmpcoo, tmpcoo2, tmpexp, tmpobj 
char    tmpsky, tmpskymef, tmpflat, tmpmsk, l_temp
struct  l_struct
real    l_skylevel, l_skysigma, l_x, l_y, l_r, l_rebin, l_mean, l_sig
real    l_expone, l_ronone, l_gainone, temp
char    l_sci_ext, l_var_ext, l_dq_ext
char    scilist, dqlist, combsig, combdq
char    tmphead, tmpdq, dqsumold, dqsum, suf, tmpstatsec
real    l_psffactor, l_psfmin, l_fwhmmin, l_skymean
bool    l_fl_vardq, l_fl_dqprop, needtoreduce

#------------------------------------------------------------------------
# Set local variables
scanfile=""
status=0
skyconst=0.
n=0
# Factor by which to multiply the PSF FWHM prior to sending to daofind.
# This will help guarantee that faint fuzzy galaxies are identified and masked
l_psffactor=2.0

l_inimages=inimages ; l_outimage=outimage
l_outtitle=outtitle
l_combtype=combtype ; l_rejtype=rejtype
l_fwhmpsf=fwhmpsf ; l_threshold=threshold
l_key_sat=key_sat ; l_key_ron=key_ron ; l_key_gain=key_gain
l_key_filter=key_filter ; l_key_airmass=key_airmass 
l_key_exptime=key_exptime ; l_masksuffix=masksuffix ; l_maskfactor=maskfactor
if( stridx(" ",l_masksuffix)>0 || stridx(",",l_masksuffix)>0 ) {
  printlog("WARNING - NISKY: Illegal masksuffix "//l_masksuffix//" changed to msk",l_logfile,verbose+)
  l_masksuffix="msk"
}
l_logfile=logfile ; l_verbose=verbose ; l_fl_gemseeing=fl_gemseeing
l_nlow=nlow ; l_nhigh=nhigh ; l_fl_keepmasks=fl_keepmasks
l_lsigma=lsigma ; l_hsigma=hsigma ; l_rebin=rebin
l_sci_ext=sci_ext ; l_dq_ext=dq_ext ; l_var_ext=var_ext
l_fl_vardq=fl_vardq ; l_fl_dqprop=fl_dqprop 
l_statsec=statsec

#------------------------------------------------------------------------
# Check for package log file or user-defined log file
cache("niri")

print(l_logfile) | scan(l_logfile)
if(l_logfile=="" || l_logfile==" ") {
   l_logfile=niri.logfile
   print(l_logfile) | scan(l_logfile)
   if(l_logfile=="" || l_logfile==" ") {
      l_logfile="niri.log"
      printlog("WARNING - NISKY:  Both nisky.logfile and niri.logfile are empty.",l_logfile, verbose+)
      printlog("                  Using default file niri.log.",l_logfile, verbose+)
   }
}
# Open log file
date | scan(l_struct)
printlog("----------------------------------------------------------------------------",l_logfile, l_verbose)
printlog("NISKY -- "//l_struct, l_logfile, l_verbose)
printlog(" ",l_logfile, l_verbose)

# Minimum permissible PSF in pixels.  This prevents possibly bad gemseeing 
# output that is dominated by bad pixels, CR, or noise from being considered.
l_psfmin=1.5
# Warn user if input PSF FWHM too small
if(l_fwhmpsf < l_psfmin) {
  printlog("WARNING - NISKY:  nisky is not designed to work on undersampled",
     l_logfile, verbose+) 
  printlog("                  data where the PSF FWHM < "//l_psfmin//" pixels.", 
     l_logfile, verbose+)
}

# check statsec
if(substr(l_statsec,1,1)!="[" || \
    substr(l_statsec,strlen(l_statsec),strlen(l_statsec))!="]" || \
    stridx(",",l_statsec)==0) {
  printlog("WARNING - NISKY:  statsec has wrong format.  Reverting to [*,*]",l_logfile,verbose+) 
  l_statsec="[*,*]"
}

# Keep imgets parameters from changing by outside world
cache("imgets")

# Make temporary file names
tmpfile1 = mktemp("tmpfl")
tmpfile2 = mktemp("tmpfl")
tmpcoo   = mktemp("tmpcoo")
tmpcoo2  = mktemp("tmpcoo2")
tmpobj   = mktemp("tmpobj")
tmpmsk   = mktemp("tmpmsk")
tmpexp   = mktemp("tmpexp")
tmpsky   = mktemp("tmpsky")
tmpskymef= mktemp("tmpskymef")
tmpflat  = mktemp("tmpflat")
tmpimg   = mktemp("tmpim")
scilist  = mktemp("tmpscilist")
dqlist   = mktemp("tmpdqlist")
dqsum    = mktemp("tmpdqsum")
dqsumold = mktemp("tmpdqsumold")
tmpdq    = mktemp("tmpdq")
combsig  = mktemp("tmpcombsig") 
combdq   = mktemp("tmpcombdq")
tmphead  = mktemp("tmphead")

#----------------------------------------------------------------------
# Start checking for valid input

# Verify that the extension names are not empty, otherwise exit gracefully
print(l_sci_ext) | scan(l_sci_ext)
if(l_sci_ext=="" || l_sci_ext==" ") {
   printlog("ERROR - NISKY: extension name sci_ext is missing.",
      l_logfile, verbose+)
   status=1
   goto clean
}

if(l_fl_vardq) {
  print(l_dq_ext) | scan(l_dq_ext)
  print(l_var_ext) | scan(l_var_ext)
  if (l_dq_ext=="" || l_dq_ext==" " || l_var_ext=="" || l_var_ext==" ") {
     printlog("WARNING - NISKY: extension name var_ext or dq_ext is missing.",
        l_logfile, verbose+)
     printlog("                 Output image will not have VAR or DQ planes.",
        l_logfile,verbose+)
     l_fl_vardq=no
     l_fl_dqprop=no
  } 
}
#----------------------------------------------------------------------
# Put all input images in a temporary file: tmpfile1

if(substr(l_inimages,1,1)=="@") {
  l_temp=substr(l_inimages,2,strlen(l_inimages))
  if(!access(l_temp)) {
    printlog("ERROR - NISKY:  Input file "//l_temp//" not found.",l_logfile,verbose+)
    status=1
    goto clean
  }
}

files(l_inimages,sort-) | unique("STDIN", > tmpfile1)

# Verify that input images actually exist. (Redundant for * though)
# at the same time check the exposure times and MEF format;
# strip .fits if present
nbad=0
nfiles=0
scanfile = tmpfile1
while(fscan(scanfile,img) != EOF) {

  # check output image
  if(nfiles==0) {
    print(l_outimage) | scan(l_outimage)
    if(l_outimage=="" || l_outimage==" ") {
        suf = substr(img,strlen(img)-4,strlen(img))
        if(suf==".fits") {
          l_outimage=substr(img,1,strlen(img)-5)//"_sky" 
        } else {
          l_outimage=img//"_sky"
        }
     }
     if(imaccess(l_outimage)) {
        printlog("ERROR - NISKY: Output file "//l_outimage//" already exists.",l_logfile,verbose+)
        status=2
        goto clean
     }
  }

  # check input images
  gimverify(img)
  if(gimverify.status==1) {
     printlog("ERROR - NISKY: File "//img//" not found.",l_logfile,verbose+)
     nbad+=1
  } else if(gimverify.status>1) {
     printlog("ERROR - NISKY: File "//img//" not a MEF FITS image.",l_logfile,verbose+)
     nbad+=1
  } else {
    imgets(img//"[0]","NPREPARE", >& "dev$null")
    if (imgets.value == "0") {
      printlog("ERROR - NISKY: Image "//img//" not NPREPAREd.",l_logfile,verbose+)
      nbad+=1
    }
  }

  # get the header of the first image and prepare for output image
  if(nfiles==0) imcopy(img//"[0]",l_outimage,verbose-)

  if (l_fl_dqprop && !imaccess(img//"["//l_dq_ext//"]") ) {
    printlog("WARNING - NISKY: Cannot propagate input DQ planes because",l_logfile,verbose+)
    printlog("                 input image "//img//" does not have a DQ plane.",l_logfile,verbose+)
    l_fl_dqprop=no
  }
  nfiles+=1
} # end while loop

# check for empty file list
if(nfiles==0) {
  printlog("ERROR - NISKY:  No input images meet wildcard criteria.",l_logfile,verbose+)
  status=1
  goto clean
}

# Exit if problems found with input files
if (nbad > 0) {
   printlog("ERROR - NISKY: "//nbad//" image(s) either do not exist, are not MEF files, or",l_logfile, verbose+)
   printlog("                 have not been run through NPREPARE.",l_logfile,verbose+)
   status=1
   goto clean
} # end if (nbad > 0)

printlog("Using input files:",l_logfile,l_verbose)
if(l_verbose) type(tmpfile1)
type(tmpfile1, >> l_logfile)
printlog("Output image: "//l_outimage,l_logfile,l_verbose)

#------------------------------------------------------------------------
# strip out science and dq planes
l_expone=0.0 ; l_ronone=0.0 ; l_gainone=0.0
n=0
scanfile = tmpfile1
while(fscan(scanfile,img) != EOF) {
   n=n+1

# strip suffix if present
   suf = substr(img,strlen(img)-3,strlen(img))
   if(substr(img,strlen(img)-4,strlen(img)) == "["//l_sci_ext//"]") {
       img=substr(img,1,(strlen(img)-5))
     }
   if(substr(img,strlen(img)-2,strlen(img)) == "[1]") {
       img=substr(img,1,(strlen(img)-3))
     }
   if(substr(img,strlen(img)-4,strlen(img)) == ".fits") {
       img=substr(img,1,(strlen(img)-5))
     }
    if(n==1) firstimage=img

# check exposure time
   imgets(img//"[0]",l_key_exptime, >>& "dev$null")
   if (imgets.value == "0") {
      printlog("ERROR - NISKY: Image header parameter not found ("//l_key_exptime//")",l_logfile, verbose+)
      status=1
      goto clean
   }
   if (n == 1) {
      l_expone=real(imgets.value)
   } else {
      if (abs(real(imgets.value)-l_expone) > 0.1) {
         printlog("WARNING - NISKY:  Exposure times are significantly different.  Continuing.",
            l_logfile, verbose+)
      }
   }

# check for gain, readnoise, and saturation values
   imgets(img//"[0]",l_key_ron, >>& "dev$null")
   if((imgets.value == "0") || (imgets.value == "")) {
      printlog("ERROR - NISKY: Could not get read noise from header of image "//img,l_logfile, verbose+)
      status=1
      goto clean
   }
   if (n == 1) {
      l_ronone=real(imgets.value)
   } else {
      if (abs(real(imgets.value)-l_ronone) > 1.) {
        printlog("WARNING - NISKY: read noise values are different.  Continuing, but the",l_logfile,verbose+)
        printlog("                 read noise in the output header will be wrong.",l_logfile,verbose+)
      }
   }

   imgets(img//"[0]",l_key_gain, >>& "dev$null")
   if((imgets.value == "0") || (imgets.value == "")) {
      printlog("ERROR - NISKY: Could not get gain from header of image"//img,l_logfile, verbose+)
      status=1
      goto clean
   }
   else l_gain=real(imgets.value)
   if (n == 1) {
      l_gainone=real(imgets.value)
   } else {
      if (abs(real(imgets.value)-l_gainone) > 0.5) {
        printlog("WARNING - NISKY: gain values are different.  Continuing, but the gain in",l_logfile,verbose+)
        printlog("                 the output header will be wrong.",l_logfile,verbose+)

      }
   }

   imgets(img//"[0]",l_key_sat, >>& "dev$null")
   if((imgets.value == "0") || (imgets.value == "")) {
      printlog("ERROR - NISKY: Could not get saturation level from header of image"//img,l_logfile, verbose+)
      status=1
      goto clean
   }

   print(img, >>tmpfile2)

# science extension
   print(img//"["//l_sci_ext//"]", >> scilist)

# DQ extension
   if(l_fl_vardq) {
     if(!imaccess(img//"["//l_dq_ext//"]")) {
       printlog("WARNING - NISKY: No DQ plane for "//img//", creating empty one.",l_logfile,verbose+)
       imarith(img//"["//l_sci_ext//"]","*","0.0",tmpdq//"_"//n//".pl",verbose-
     } else {
       imcopy(img//"["//l_dq_ext//"]",tmpdq//"_"//n//".pl",verbose-)
     }
     if (l_fl_dqprop) {
        print(tmpdq//"_"//n//".pl", >> dqlist) 
     }
     hedit(img//"["//l_sci_ext//"]","BPM",tmpdq//"_"//n//".pl",add+,upd+,ver-,show-)
     
   }
} # end of while loop
scanfile=""

l_ronone = l_ronone/sqrt(n)
l_ronone = real(int(l_ronone*10.))/10.
l_gainone = l_gainone*n

if (n == 1) {
   printlog("ERROR - NISKY: Cannot combine a single image.", 
      l_logfile, verbose+)
   status=1
   goto clean
} else if (n == 0) {
   printlog("ERROR - NISKY: No images to combine.",l_logfile,verbose+)
   status=1
   goto clean
}

#---------------------------------------------------------------------------
# Check for mask images that already exist

needtoreduce=no
i=1
scanfile=tmpfile2
while(fscan(scanfile,img) != EOF) { 
  if(!access(img//l_masksuffix//".pl")) { 
    needtoreduce=yes
  }
}
scanfile=""

#-----------------------------------------------------------------------------
# Make a quick temporary sky image using nifastsky.cl and flat field

if(needtoreduce) {
  printlog("NISKY calling NIFASTSKY",l_logfile,l_verbose)
  nifastsky("@"//tmpfile2,outimage=tmpskymef,outtitle="default",combtype="default",
     sci_ext=l_sci_ext,var_ext="",dq_ext="",fl_vardq=no,
     fl_dqprop=no,key_exptime=l_key_exptime,verbose=l_verbose,logfile=l_logfile)
  if (nifastsky.status != 0) {
    printlog("ERROR - NISKY: Error in NIFASTSKY prevents NISKY completion.",
      l_logfile,verbose+)
    status=1
    goto clean
  }
  printlog("Returning to NISKY",l_logfile,l_verbose)

  imcopy(tmpskymef//"["//l_sci_ext//"]",tmpsky,verbose-)
  imstat(tmpsky//l_statsec,fi="midpt,stddev",format-,binwidth=0.05,lower=INDEF,
     upper=INDEF) | scan(l_mean,l_sig)
  imstat(tmpsky//l_statsec,fi="mean",upper=(l_mean+4*l_sig),lower=(l_mean-4*l_sig),
     binwidth=0.05) | scan(l_mean)
  if(l_mean==INDEF) {
    printlog("ERROR - NISKY: Statistics failed, possibly due to a bad statsec.",l_logfile,verbose+)
    status=1
    goto clean
  }

  l_skymean=l_mean
  if(l_mean <= 0.) l_mean=0.0001
  if(sqrt(1./(l_mean*n*l_gain))<=0.02) {
    l_temp="((a>"//(l_mean+4.*l_sig)//") || (a<"//(l_mean-4.*l_sig)//")) ? 1. : (a/"//l_mean//")"
    imexpr(l_temp,tmpflat,tmpsky,outtype="real", >>& "dev$null")
  } else {
    imexpr("a*0.+1.",tmpflat,tmpsky,outtype="real", >>& "dev$null")
  }

# set the image size variables
  imgets(tmpflat,"i_naxis1", >>& "dev$null")
  l_xsize = int(imgets.value)
  imgets(tmpflat,"i_naxis2", >>& "dev$null")
  l_ysize = int(imgets.value)
} # end if(needtoreduce)

#---------------------------------------------------------------------------
# Reduce the images, identify objects, and make a mask for each input file
# (if the object mask file already exists then that file is used and task
# continues)

cache("tinfo")
scanfile=tmpfile2

# Start mask-making loop here
i=1
while(fscan(scanfile,img) != EOF) { 

imgsci=img//"["//l_sci_ext//"]"
if(l_fl_vardq) imgdq=tmpdq//"_"//i//".pl"

# Check for existing masks
if(!access(img//l_masksuffix//".pl")) { 

  printlog("Making mask for image "//img,l_logfile,l_verbose)

# Reduce the input image using the fast sky and flat constructed above
  imstat(imgsci//l_statsec,fi="midpt,stddev",binwidth=0.05,
     lower=INDEF,upper=INDEF,format-) | scan(l_mean,l_sig)
  imstat(imgsci//l_statsec,fi="midpt",binwidth=0.05,
     lower=(l_mean-4.*l_sig),upper=(l_mean+4.*l_sig),format-) | scan(skyconst)
  if(skyconst==INDEF) {
    printlog("ERROR - NISKY: Statistics failed, possibly due to a bad statsec.",l_logfile,verbose+
    status=1
    goto clean
  }
  imexpr("((a-(b*"//(skyconst/l_skymean)//"))/c)+"//skyconst,
     tmpimg,imgsci,tmpsky,tmpflat,rangecheck=yes,outtype="real", 
     >>& "dev$null")
  if(l_fl_vardq) {
    nmisc.fixpix(tmpimg,imgdq,linterp="",cinterp="",ninterp="1",creplace="",>>& "dev$null")
  }

# check gain, readnoise, and saturation values for each image
  imgets(img//"[0]",l_key_ron)
  l_ron=real(imgets.value)*sqrt(2.)
# increase the noise by sqrt(2) to account for sky subtraction

  imgets(img//"[0]",l_key_gain)
  l_gain=real(imgets.value)

  imgets(img//"[0]",l_key_sat)
  l_datamax=real(imgets.value)

# Rebin, if FWHM > (rebin) pix, to speed up the daofind part
  
  if(l_fwhmpsf > l_rebin) {
     l_binfactor = nint(l_rebin/2.)
     printlog("Binning by a factor of "//l_binfactor,l_logfile,l_verbose)
     blkavg(tmpimg,tmpimg,l_binfactor,l_binfactor,option="average")
     l_fwhmpsf = l_fwhmpsf/l_binfactor
     l_ron = l_ron*l_binfactor
     l_gain = l_gain*l_binfactor*l_binfactor

     #fix statsec for rebinned image
     l_temp=substr(l_statsec,2,(stridx(",",l_statsec)-1))
     if (l_temp!="*") {
       if(stridx(":",l_temp)>0) {
         x1=nint(real(substr(l_temp,1,(stridx(":",l_temp)-1))) / l_binfactor)
         x2=nint(real(substr(l_temp,(stridx(":",l_temp)+1),strlen(l_temp))) / l_binfactor)    
         l_temp=x1//":"//x2
       } else {
         x1=nint(real(l_temp) / l_binfactor)
         l_temp=x1
       }
     }
     tmpstatsec="["//l_temp//","
     l_temp=substr(l_statsec,(stridx(",",l_statsec)+1),(stridx("]",l_statsec)-1))
     if (l_temp!="*") {
       if(stridx(":",l_temp)>0) {
         y1=nint(real(substr(l_temp,1,(stridx(":",l_temp)-1))) / l_binfactor)
         y2=nint(real(substr(l_temp,(stridx(":",l_temp)+1),strlen(l_temp))) / l_binfactor)   
         l_temp=y1//":"//y2
       } else {
         y1=nint(real(l_temp) / l_binfactor)
         l_temp=y1
       }

     }
     tmpstatsec=tmpstatsec//l_temp//"]"
  }
  else {
    l_binfactor=1
    tmpstatsec=l_statsec
  }

# Get some header keywords from input into tmpimg, for gemseeing and daofind
  imgets(img//"[0]",l_key_airmass, >>& "dev$null")
  if((imgets.value != "0") && (imgets.value != "")) {
     temp=real(imgets.value)
     hedit(tmpimg,l_key_airmass,temp,add+,ver-,upd+, >>& "dev$null")
  }
  else {
     l_temp="INDEF"
     hedit(tmpimg,l_key_airmass,l_temp,add+,ver-,upd+, >>& "dev$null")
  } 
  imgets(img//"[0]",l_key_filter, >>& "dev$null")
  if((imgets.value != "0") && (imgets.value != "")) {
     l_temp=imgets.value
     hedit(tmpimg,l_key_filter,l_temp,add+,ver-,upd+, >>& "dev$null")
  }
  else {
     l_temp="INDEF"
     hedit(tmpimg,l_key_filter,l_temp,add+,ver-,upd+, >>& "dev$null")
  } 
  imgets(img//"[0]","CAMERA", >>& "dev$null")
  if((imgets.value != "0") && (imgets.value != "")) {
     l_temp=imgets.value
     hedit(tmpimg,"CAMERA",l_temp,add+,ver-,upd+, >>& "dev$null")
  } 
  else {
     l_temp="INDEF"
     hedit(tmpimg,"CAMERA",l_temp,add+,ver-,upd+, >>& "dev$null")
  } 
  imgets(img//"[0]","INSTRUME", >>& "dev$null")
  if((imgets.value != "0") && (imgets.value != "")) l_temp=imgets.value
  else l_temp="INDEF"
  hedit(tmpimg,"INSTRUME",l_temp,add+,ver-,upd+, >>& "dev$null")

#Determine the seeing using GEMSEEING

  if(l_fl_gemseeing) {
    l_fwhmmin=max(l_psfmin,l_fwhmpsf/4.)

    if(i==1) {
       printlog("Occasionally a bug in the binary IRAF code will lead to an invalid floating",l_logfile,l_verbose)
       printlog("point operation during the execution of GEMSEEING.  If NISKY appears inactive",l_logfile,l_verbose)
       printlog("for more than about a minute after calling GEMSEEING, please stop the script",l_logfile,l_verbose)
       printlog("using CTRL-C and restart NISKY with fl_gemseeing=no or check to make sure",l_logfile,l_verbose)
       printlog("that the fwhmpsf parameter is set appropriately.",l_logfile,l_verbose)
       printlog(" ",l_logfile,l_verbose)
    }
    printlog("Calling GEMSEEING.",l_logfile,l_verbose)
    gemseeing(tmpimg,fl_keep=no,fl_update=yes,psffile="default",
       fl_useall-,maxnobj=100,fl_overwrite+,sigthres=l_threshold,
       threshold=INDEF,datamax=l_datamax,
       ron=l_ron,gain=l_gain,key_ron="",key_gain="",
       fwhmin=l_fwhmpsf,width=4.,mdaofaint=0.,
       fwhmmin=l_fwhmmin,fwhmmax=0.,verbose-,fl_database=no,
       fl_strehl-,fl_cleed+,fl_inter-,instrument="",camera="",
       key_inst="INSTRUME",key_camera="CAMERA",key_filter=l_key_filter,
       sharplow=0.3,sharphigh=1,sharpsig=3,dmagsig=2,fwhmsig=2.5,
       pixscale=1.,fl_inter=no,logfile="", >>& "dev$null")

    imgets(tmpimg,"FWHMPSF",>>& "/dev/null")
    if(real(imgets.value) == 0.) {       
       printlog("WARNING - NISKY:  GEMSEEING failed to find any objects in "//img,l_logfile,verbose+)
       printlog("                  Staying with previous PSF FWHM value of "//l_fwhmpsf//" pixels.",l_logfile,verbose+)
    } 
    else if (real(imgets.value) > l_fwhmmin) {
       l_fwhmpsf = real(imgets.value)
       printf("Adopting new PSF FWHM value from GEMSEEING of %4.2f  pix\n",(l_fwhmpsf*l_binfactor)) | scan(l_struct)
        printlog(l_struct,l_logfile,l_verbose)
        printlog(" ",l_logfile,l_verbose)
    } else {
       printlog("WARNING - NISKY:  GEMSEEING returned a PSF FWHM value that was apparently",l_logfile,verbose+)
       printlog("                  too small.  Staying with the previous PSF FWHM value.",l_logfile,verbose+)
    }
  }

# Identify stars in reduced sky images using daofind

# soften the PSF estimate to help find and remove faint fuzzies
  l_fwhmpsf = l_fwhmpsf * l_psffactor

# Get a reasonable estimate of the noise in the background,
  imstat(tmpimg//tmpstatsec,fields="stddev,midpt",lower=INDEF,
     upper=INDEF,binwidth=0.05,format-) | scan(l_skysigma,l_skylevel)
  imstat(tmpimg//tmpstatsec,fields="stddev,midpt",
     lower=(l_skylevel-4*l_skysigma),upper=(l_skylevel+4*l_skysigma),
     binwidth=0.05,format-) | scan(l_skysigma,l_skylevel)
  if(l_skysigma==INDEF) {
    printlog("ERROR - NISKY: Statistics failed, possibly due to a bad statsec.",l_logfile,verbose+
    status=1
    goto clean
  }
  
  l_fl_objects=yes
  daofind (tmpimg,output=tmpimg//".coo",datapars="", 
    scale=1,fwhmpsf=l_fwhmpsf,threshold=l_threshold,
    sigma=l_skysigma,ratio=1.,theta=0.,nsigma=1.5,sharplo=0.,
    sharphi=100.,roundlo=-100.,roundhi=100.,boundary="nearest", 
    constant=0.,graphics="stdgraph",display="stdimage",icommands="", 
    gcommands="",mkdetections=no,interactive=no,verify=no,verbose-,
    emission=yes,datamin=0.0,datamax=l_datamax,noise="poisson",
    ccdread="",gain="",readnoise=l_ron,epadu=l_gain,filter=l_key_filter,
    airmass=l_key_airmass)

# put the PSF FWHM back 
  l_fwhmpsf = l_fwhmpsf / l_psffactor

# classify these "objects" into objects and cosmic-ray-hits
# and assign radii
# CRs    : radius=1        sharpness outside 0.2-1.0 
#                       OR round outside -1 -> 1.
# objects: radius ....
# Based on an empirical formula developed by IJ and including a factor of 1.5
# added by JJ.  Additional scaling applied via maskfactor.
#   
# Make the coordinate file an STSDAS table 
  pconvert(tmpimg//".coo",tmpcoo,"*",expr=yes,append-)
  tinfo(tmpcoo, >>& "dev$null")
  if(tinfo.nrows < 1) l_fl_objects=no

  if(l_fl_objects) {
    tcalc(tmpcoo,"peak","10**(-mag/2.5)*"//str(l_threshold*l_skysigma),
      colfmt="f6.0",colunit="")
    tcalc(tmpcoo,"peak",
      "if sharpness/0.7>1 then max(sharpness/0.7,0.3)*peak else peak")
    tcalc(tmpcoo,"radius",
      "(1./(2**0.25-1.)) * ((peak/"//str(l_skysigma)//")**0.25-1.)",
      colfmt="f6.3",colunit="")
    tcalc(tmpcoo,"radius",
      "if radius>0 then sqrt(radius)*"//l_fwhmpsf//"*1.40/2. else 0.5")
    tcalc(tmpcoo,"radius",
      "if radius>1 then "//l_maskfactor//"*1.5*radius else radius")

# objects with peak above l_datamax and low sharp are mostly
# saturated stars, make radius for these larger
    tcalc(tmpcoo,"radius",
      "if sharpness<0.6 && peak>="//l_datamax//" then radius*1.8 else radius")
# set radii of CRs
    tcalc(tmpcoo,"radius",
      "if sharpness<0.2 || sharpness>1 then 0.5 else radius",
      colfmt="f6.3",colunit="")
    tcalc(tmpcoo,"radius","if ground<-1 || ground>1 then 0.5 else radius",
      colfmt="f6.3",colunit="")

# Get back into the input coordinate frame
    tcalc(tmpcoo,"radius","radius*"//l_binfactor)
    tcalc(tmpcoo,"xcenter","xcenter*"//l_binfactor)
    tcalc(tmpcoo,"ycenter","ycenter*"//l_binfactor)
    l_ron = l_ron/l_binfactor
    l_gain = l_gain/(l_binfactor*l_binfactor)
    l_fwhmpsf = l_fwhmpsf*l_binfactor

# Check for INDEFs in the mask file (may want to replace with 
    tselect(tmpcoo,tmpcoo2,"radius > 0.1")

    tprint(tmpcoo2,col="xcenter,ycenter,radius",showrow-,showhdr-,
       showunits-, > tmpexp)
  } # end if(l_fl_objects)
  else { # make an empty mask
    print (" ", > tmpexp)
  }

# make the mask
  skymask(tmpexp,refimages="",masks=img//l_masksuffix//".pl",
     defshape="point",ncols=l_xsize,nlines=l_ysize,badvalue=1,goodvalue=0,
     verbose-)

# Include the original DQ plane
   if(l_fl_vardq) {
     imarith(imgdq,"max",img//l_masksuffix//".pl",img//l_masksuffix//".pl",pixtype="short",verbose-)
   }

} else { 
    printlog("WARNING - NISKY: Using existing mask file "//img//l_masksuffix//".pl",l_logfile,verbose+)
  } # end if(!access(img//l_masksuffix//".pl"))

# edit header to include mask file as BPM
  hedit(img//"["//l_sci_ext//"]","BPM",img//l_masksuffix//".pl",add+,ver-,update+,show-)
  if(l_fl_keepmasks) {  
    gemhedit(img//"[0]","NISKYMSK",img//l_masksuffix//".pl","Object mask generated by NISKY")
  }

# get rid of the temporary reduced image and the coordinate files
  delete(tmpexp,verify-, >>& "/dev/null")
  delete(tmpobj//".tab",verify-, >>& "/dev/null")
  delete(tmpcoo//".tab",verify-, >>& "/dev/null")
  delete(tmpcoo2//".tab",verify-, >>& "/dev/null")
  delete(tmpimg//".coo",verify-, >>& "/dev/null")
  delete(tmpimg//"_coo",verify-, >>& "/dev/null")
  delete(tmpimg//"_see",verify-, >>& "/dev/null")
  delete(tmpimg//"_see.tab",verify-, >>& "/dev/null")
  imdelete(tmpimg,verify-, >>& "/dev/null")

  i+=1
} # end while loop
scanfile=""
 
#---------------------------------------------------------------------
# imcombine raw sky frames with star masks as bad pixel masks

# save the user's parameters for imcombine
delete("uparm$imhimcome.par.org",ver-, >>& "/dev/null")
if(access("uparm$imhimcome.par"))
  copy("uparm$imhimcome.par","uparm$imhimcome.par.org",verbose-)

cache("imcombine")

# Set imcombine parameters 
imcombine.sigma = combsig
imcombine.plfile = combdq
imcombine.logfile = l_logfile
imcombine.project = no
imcombine.outtype = "real"
imcombine.offsets = "none"
imcombine.masktype = "goodvalue"
imcombine.maskvalue = 0
imcombine.blank = 0.
imcombine.scale = "median"
imcombine.zero = "none"
imcombine.weight = "none"
imcombine.statsec = l_statsec
imcombine.expname = ""
imcombine.lthreshold = INDEF
imcombine.hthreshold = INDEF
imcombine.nkeep = 1
imcombine.grow = 0
imcombine.mclip=yes
imcombine.sigscale=0.1

if(l_combtype=="default") {
  imcombine.combine = "average"
  if (n == 2) {
    imcombine.reject = "minmax"
    imcombine.nlow=0
    imcombine.nhigh=1
    printlog("WARNING - NISKY: Combining two images by taking the minimum.",
       l_logfile,verbose+)
    } 
  else {
    imcombine.reject = "avsigclip"
    imcombine.lsigma = 3.
    imcombine.hsigma = 3.
    }
  } 
else { 
  imcombine.combine = l_combtype
  imcombine.reject= l_rejtype
  imcombine.nlow = l_nlow
  imcombine.nhigh = l_nhigh
  imcombine.lsigma = l_lsigma
  imcombine.hsigma = l_hsigma
  if (n < 5) {
    printlog("WARNING - NISKY: Combining 4 or fewer images using "//l_combtype,l_logfile,verbose+)
    if(l_rejtype == "minmax") {
        printlog("                      with "//l_nlow//" low and "//l_nhigh//" high pixels rejected.", l_logfile,verbose+)
     } else if(l_rejtype == "avsigclip") {
        printlog("                  with "//l_lsigma//"=lower sigma and "//l_hsigma//"=upper sigma.",l_logfile,verbose+)

     } else {
       printlog("                   with no pixels rejected.",l_logfile,verbose+)
     }
  }
  if((n <= (l_nlow+l_nhigh)) && (l_rejtype=="minmax")) {
     printlog("ERROR - NISKY: Cannot reject more pixels than the number of images.",l_logfile,verbose+)
     status=1
     goto clean
  }

} #end not-default section

# Do the combining (supress the output from imcombine)
imcombine("@"//scilist,l_outimage//"["//l_sci_ext//",append]")

printlog("Combining "//str(n)//" images, using "//imcombine.combine,
   l_logfile,l_verbose)
printlog("Rejection type is "//imcombine.reject,
   l_logfile,l_verbose)
if(imcombine.reject=="minmax") {
  printlog("with "//imcombine.nlow//" low and "//imcombine.nhigh//" high values rejected.",
     l_logfile,l_verbose)
}
if(imcombine.reject=="avsigclip") {
  printlog("with lsigma="//imcombine.lsigma//" and hsigma="//imcombine.hsigma,
     l_logfile,l_verbose)
}

if(l_fl_vardq) {
# make variance image by squaring combsig
   imarith(combsig,"*",combsig,l_outimage//"["//l_var_ext//",append]",
      pixtype="real")
# bad pixels have val=n, or propagate input DQ values if requested
   if(!l_fl_dqprop) {
      imexpr("(a=="//n//") ? 1 : 0",l_outimage//"["//l_dq_ext//",append]",
         combdq,outtype="short", >>& "dev$null")
   }
   else {
      scanfile=dqlist
      for (j=1;j<=n;j+=1) {
         i=fscan(scanfile, img)
         imarith(img,"*",combdq,img,pixtype="real",verbose-)
         if (j==1) {
            imcopy(combdq,dqsumold,verbose-)
         } else {
            imdelete(dqsumold,verify-, >& "dev$null")
            imrename(dqsum,dqsumold,verbose-)
         }
         addmasks(dqsumold//","//img,dqsum//".pl","im1 || im2")
         imcopy(dqsum,l_outimage//"["//l_dq_ext//",append]", >& "dev$null")
      } #end for
      scanfile=""
   }

# update headers
   hedit(l_outimage//"["//l_sci_ext//"]","BPM","",delete+,verify-,upd+,show-)
   hedit(l_outimage//"["//l_var_ext//"]","BPM","",delete+,verify-,upd+,show-)
   hedit(l_outimage//"["//l_dq_ext//"]","BPM","",delete+,verify-,upd+,show-)

} # end if(l_fl_vardq)

#--------------------------------------------------------------------------
# update headers in final image PHU

date | scan(l_struct)
gemhedit(l_outimage//"[0]","GEM-TLM",l_struct,"Last modification with GEMINI")
printf("%-8s= \'%-18s\' / %-s\n","NISKY",l_struct,"Time stamp for NISKY", >> tmphead)

if(l_outtitle=="default" || l_outtitle=="" || l_outtitle==" ")
  gemhedit(l_outimage//"[0]","i_title","SKY IMAGE from gemini.niri.nisky","Image title")
else
  gemhedit(l_outimage//"[0]","i_title",l_outtitle,"Image title")

l_ronone=(int(l_ronone*10000.))/10000.
l_gainone=(int(l_gainone*10.))/10.
hedit(l_outimage//"[0]",l_key_ron,l_ronone,upd+,ver-,show-)
hedit(l_outimage//"[0]",l_key_gain,l_gainone,upd+,ver-,show-)

printf("%-8s= %20.2f / %-s\n","NISKYMSF",l_maskfactor,"Mask scale factor used by NISKY", >> tmphead)
printf("%-8s= %20.0f / %-s\n","NISKYREB",l_rebin,"Rebin size used by NISKY", >> tmphead)
printf("%-8s= \'%-18s\' / %-s\n","NISKYSTA",l_statsec,"Statistics region used by NISKY", >> tmphead)

if(!l_fl_gemseeing) {
  printf("%-8s= %20.2f / %-s\n","NISKYFWH",l_fwhmpsf,"PSF FWHM used by NISKY", >> tmphead)
} else {
  printf("%-8s= %20.2f / %-s\n","NISKYFWH",l_fwhmpsf,"PSF FWHM in the last image", >> tmphead)
}

printf("%-8s= \'%-18s\' / %-s\n","NISKYCOM",imcombine.combine,"Type of combine used by NISKY", >> tmphead)
printf("%-8s= \'%-18s\' / %-s\n","NISKYREJ",imcombine.reject,"Type of rejection used by NISKY", >> tmphead)
if(imcombine.reject=="minmax") {
  printf("%-8s= %20.0f / %-s\n","NISKYNLO",imcombine.nlow,"Low pixels rejected (minmax)", >> tmphead)
  printf("%-8s= %20.0f / %-s\n","NISKYNHI",imcombine.nhigh,"High pixels rejected (minmax)", >> tmphead)
}
if(imcombine.reject=="avsigclip") {
  printf("%-8s= %20.2f / %-s\n","NISKYLSI",imcombine.lsigma,"Lower sigma for rejection", >> tmphead)
  printf("%-8s= %20.2f / %-s\n","NISKYHSI",imcombine.hsigma,"Upper sigma for rejection", >> tmphead)
  printf("%-8s= %20.0f / %-s\n","NISKYNKE",imcombine.nkeep,"Min(max) number of pixels to keep(reject)", >> tmphead)
  printf("%-8s= \'%-18b\' / %-s\n","NISKYMCL",imcombine.mclip,"Use median in clipping algorithms", >> tmphead)
  printf("%-8s= %20.2f / %-s\n","NISKYSSC",imcombine.sigscale,"Tolerance for sigma clip scaling correction", >> tmphead)
}
printf("%-8s= \'%-18b\' / %-s\n","NISKYDQP",l_fl_dqprop,"Retain input data quality planes?", >> tmphead)

# Put input image names in header
i=1
scanfile=tmpfile1
while(fscan(scanfile,img) != EOF) {
  printf("%-8s= \'%-18s\' / %-s\n","NISKIM"//str(i),img,"Input image combined with NISKY", >> tmphead)
  i=i+1
}

# put all the new stuff in the header
mkheader(l_outimage//"[0]",tmphead,append+,verbose-)

scanfile=""

#-------------------------------------------------------------------------
clean:
  if(status==0) {
    printlog("NISKY exit status:  good.",l_logfile,l_verbose)
  } else {
    imdelete(l_outimage,ver-, >& "dev$null")
  }
  printlog("----------------------------------------------------------------------------",l_logfile,l_verbose)

   if(access(tmpfile2)) {
      scanfile=tmpfile2
      while(fscan(scanfile,img) != EOF) {
        hedit(img//"["//l_sci_ext//"]","BPM","",delete+,verify-,upd+,show-)
        if(!l_fl_keepmasks) {
           imdelete(img//l_masksuffix//".pl",verify-, >>& "/dev/null")
           hedit(img//"[0]","NISKYMSK",del+,upd+,ver-, >>& "/dev/null")
        }
      }
   }

   delete(tmpfile1,ver-, >>& "/dev/null")
   delete(tmpfile2,ver-, >>& "/dev/null")
   delete(scilist//","//dqlist//","//tmphead,verify-, >& "dev$null")
   imdelete(combsig//","//combdq,verify-, >& "dev$null")
   imdelete(dqsum//","//dqsumold,verify-, >& "dev$null")
   imdelete(tmpdq//"*.pl",verify-, >& "dev$null")
   imdelete(tmpsky,ver-, >>& "/dev/null")
   imdelete(tmpskymef,ver-, >>& "/dev/null")
   imdelete(tmpflat,ver-, >>& "/dev/null")
   unlearn("imcombine")
# restore the user's parameters for imcombine
   if(access("uparm$imhimcome.par.org"))
      rename("uparm$imhimcome.par.org","uparm$imhimcome.par",field="all")
   
end
